.CT
Code Generator Operation
.LP
The code generator recursively descends over the parse tree, transforming
it into code.  The most central function of the code generator is the 
translation of arithmetic expressions, and this function is discussed first.
Next, the code generation technique for boolean expressions is
explained, and finally, code generation at the statement level is
demonstrated with examples of how it ties in with the other levels.
.NH 0
Expression Code Generation
.NH 2
The Expression Stack
.LP
Expression translation is done, conceptually, by a postfix traversal of the
tree, which outputs operations in the order a stack machine could process
them.  Unfortunately, the target machine is not as simple as a stack machine.
Thus the ``stack'' holding partially evaluated expressions is kept at compile
time.  We have called this the ``expression stack''.  This is described in the
following section.
.KS
.LP
The expression stack starts out empty.  When an expression is evaluated, it
holds the result, just as the stack would in a stack machine.  The difference
is that the expression stack holds only the information required to obtain
the result, not the result itself.  In tcc, this information takes the form
of a VAX addressing mode.  Each stack entry is described by a 
\f(CWtype\fP
field and two \f(CWvalue\fP
fields, which we will denote as 
\f(CWx\fP and \f(CWy\fP.  Then the
expression stack can hold the following:
.sp
.TS
center;
l l l.
Name	DEC Notation	Unix Notation
_
\f(CWES_REG	Rx	rx
ES_REG+ES_DEF	(Rx)	(rx)
ES_CONST	#x	$x
ES_POSTINCR	(Rx)+	(rx)+
ES_POSTINCR+ES_DEF	@(Rx)+	*(rx)+
ES_PREDECR	-(Rx)	-(rx)
ES_DREGOFFSET	y(Rx)	y(rx)
ES_DREGOFFSET+ES_DEF	@y(Rx)	*y(rx)
ES_LABELADR	#_label+y	$_label+y
ES_STATICADR	#Lx+y	$Lx+y
ES_LABEL	_label+y	_label+y
ES_LABEL+ES_DEF	@_label+y	*_label+y
ES_STATIC	Lx+y	Lx+y
ES_STATIC+ES_DEF	@Lx+y	*Lx+y\fP
_
.TE
.KE
.sp
.KS
.LP
The code generator begins at a leaf node of an expression by pushing onto the
expression stack an addressing mode corresponding to the object there.  The
following table illustrates what each kind of addressing mode is pushed for
various objects:
.sp
.TS
center;
lw(4i) l.
Type of object	Addressing Mode
_
constant expression evaluating to \f(CWx\fP	\f(CW#x\fP

global variable at label \f(CW_name\fP	\f(CW_name+0\fP

T{
local variable at offset \f(CWy\fP from frame pointer
T}	\f(CWy(R13)\fP

T{
function parameter at offset \f(CWy\fP from argument pointer
T}	\f(CWy(R12)\fP

quoted string (stored at Lx)	\f(CWLx+0\fP

local static variable	\f(CWLx+0\fP
_
.TE
.KE
.sp
.KS
.LP
Now as the code generator proceeds up the tree, it attempts to modify the
addressing modes on the expression stack to represent the effect of the
operator.  Code is generated only when this is not possible to do.  There are
three major functions for manipulating the expression stack entries.
They correspond to C operators as follows:
.sp
.TS
center;
c c
c l.
C operator	Function name
_
\f(CW.	addoffset()
*	indirect()
&	addressof()\fP
_
.TE
.KE
.sp
.KS
.LP
As an example, suppose the top entry on the expression stack is 
\f(CW_label+0\fP,
and a dot operator is applied to it.  It is found that the structure member
named after the dot is at offset 16 into the structure.  Thus the stack
entry is simply transformed into 
\f(CW_label+16\fP.  Or suppose that the
expression stack contains 
\f(CWR0\fP and two \f(CW*\fP
operators in succession are applied.
The stack entry would become 
\f(CW@(R0)\fP.  On the other hand, suppose that the
stack entry already contains 
\f(CW@-16(R12)\fP, and another \f(CW*\fP
operator is applied.
There is no addressing mode which can represent this, so code must be
generated.  The \f(CWindirect()\fP function will now allocate a register,
say 
\f(CWR0\fP, and generate this instruction:
.DS
\f(CWMOVL    @-16(R12),R0
.DE
.KE
.KS
.LP
Then it will replace the top entry on the expression stack with 
\f(CW(R0)\fP.  It
should be noted that the code generator always attempts to have an ``effective
address'' on the stack rather than a value, until this is no longer possible.
This is to reserve the possibility of applying an \f(CW&\fP
operator to the entry
or to store something in it, which cannot be done once the entry is in a
register.  As an illustration, suppose a stack entry is 
\f(CW@(R0)\fP and three \f(CW*\fP
operators are applied to it in succession.  The code generator will generate
this code:
.DS
\f(CWMOVL    @(R0),R0            ; stack contains (R0)
MOVL    (R0),R0             ; still (R0)
MOVL    (R0),R0             ; still (R0)
.DE
.KE
.KS
.LP
And end up with \f(CW(R0)\fP on the stack.  Note that it would have been more 
efficient to generate this code:
.DS
\f(CWMOVL    @(R0),R0
MOVL    @(R0),R0
.DE
.KE
.LP
And end up with \f(CWR0\fP
on the stack.  However, the value would now be in a
register rather than the address, and if the next thing to come along were an 
\f(CW&\fP or assignment operator, it could not be done.  Therefore, the code
generator, which as described so far cannot look ``up'' the tree to see if
the address of an item will be required, has to sacrifice an occasional
instruction to be on the safe side.  Fortunately, constructs such as 
\f(CW***a\fP in C are very rare and thus this is not a big shortcoming.
.KS
.NH 2
The genexpr() Function
.LP
The heart of the expression translator in tcc is a function called 
\f(CWgenexpr()\fP.  It takes three arguments, as shown:
.DS
\f(CWint genexpr(node,rp,maxsize)struct treenode *node;
struct es_entry *rp;
int maxsize;
.DE
.KE
.KS
.LP
The simplest invocation has the 
\f(CWrp\fP field set to zero and the \f(CWmaxsize\fP
field set to 4 (the size, in bytes, of an integer or pointer).  Then the
genexpr() simply translates the entire parse tree rooted at 
\f(CWnode\fP (which
must be the tree for an expression), and returns with an extra value on the
expression stack referring to result.  Suppose the expression 
\f(CWa+1\fP were
to be translated, and 
\f(CWnode\fP pointed to the tree node containing the \f(CW+\fP
operation.  Then genexpr would generate the following code:
.DS
\f(CWADDL3   _a,#1,R0
.DE
.KE
.LP
And return \f(CWR0\fP
on the expression stack.  Note that R0 is a value, not an
effective address, but this is permissible and unavoidable since an expression
like \f(CWa+1\fP
can neither be taken the address of nor assigned to.  It would, in
fact, be a fatal internal error in the compiler if the type checker/semantic
analyzer permitted such an operation at this point.
.KS
.LP
At this point, the \f(CWrp\fP
argument to genexpr(), and the value it returns,
should be discussed.  These permit a simple lookahead mechanism which improves
code generation.  To illustrate why they are needed, suppose the expression
\f(CWa\ =\ b\ +\ c\fP
must be translated.  Without the lookahead mechanism, the best
the code generator could do is this:
.DS
\f(CWADDL3   _a,_b,R0
MOVL    R0,_c
.DE
.KE
.KS
.LP
It turns out that in the VAX, the result of an operation can often be stored
in memory ``for free'' by just using a more complex destination field.  That is,
we wish to generate this instruction:
.DS
\f(CWADDL3   _a,_b,_c
.DE
.KE
.KS
.LP
This is accomplished using the ``advance knowledge'' the code generator has
of the operations to be performed, since it processes the tree from the top
down.  Thus, when the \f(CW=\fP
node is seen, the code generator already knows that
the result of genexpr() for \f(CWa+b\fP
will be stored in \f(CWc\fP.  This is where the \f(CWrp\fP
(for ``result pointer'') mechanism comes in.  The code generator simply
evaluates the left side of the 
\f(CW=\fP node first, resulting in an expression stack entry of 
\f(CW_c\fP.  Now it calls the right side, with 
\f(CWrp\fP pointing to the stack entry of 
\f(CW_c\fP.  This instructs the genexpr() function that if convenient,
it can deposit the result directly there and not generate an expression
stack entry at all.  If this is possible, genexpr() will do it and return
non-zero.  If not possible, it will deposit the result on the stack as
before and return zero.  Here is another example where this is convenient.
Suppose the result of \f(CWa+b\fP
must be pushed on the stack as a function argument.
Therefore, the caller to genexpr() constructs an expression stack entry of
this form: \f(CW-(R14)\fP and points \f(CWrp\fP
to it.  Now genexpr() obligingly uses this
as the destination, and generates this instruction:
.DS
\f(CWADDL3   _a,_b,-(R14)
.DE
.KE
.KS
.LP
Suppose on the other hand that genexpr() did not find it convenient to 
deposit the result directly.  Assume the operation was ``push *a on the stack''.
\f(CW*a\fP
can be readily represented as an addressing mode, and so genexpr() will
represent it on the expression stack as \f(CW@_a\fP without generating any code
at all.  It signals this by returning zero.  Now the caller has to pop the
entry off the stack and generate its own instruction, like this:
.DS
\f(CWPUSHL   @_a
.DE
.KE
.LP
The only other argument to genexpr() is the \f(CWmaxsize\fP
parameter.  This exists
because not every result expected from genexpr() is 32 bits long.  Some are
shorter.  \f(CWmaxsize\fP serves two purposes.  For one, it says that genexpr()
need not bother evaluating the expression to any greater precision, and for
the other, it says that when \f(CWrp\fP
is set, genexpr() may only store a value exactly that big.
.KS
.NH 2
Extension of Short Values to Longer Ones
.LP
The definition of the C language states that short types must be ``integer
promoted'' before practically anything can be done with them.  The type
checker takes care of this by inserting \f(CWextend\fP nodes into the parse tree
where necessary.  Suppose a and c are character variables, and b is an
unsigned short variable, and the operation 
\f(CWa = b + c\fP
is performed.  The
type checker would treat this like this:
.DS
\f(CWa = promote(b) + promote(c)
.DE
.KE
.LP
Where promote(b) would be a node for ``zero fill upper 16 bits'' and promote(c)
would be a node for ``sign extend upper 24 bits''.  The addition would then be
performed at the 32-bit (i.e. integer) level, as specified by the C standard,
and the result would be truncated to fit into the 8-bit variable \f(CWa\fP.
The code generator would see the \f(CW=\fP
node first, and realize that the result
need only be eight bits, and can be directly stored at location 
\f(CW_a\fP.  It would thus call genexpr() with \f(CWrp\fP set to the entry for 
\f(CW_a\fP, and \f(CWmaxsize\fP set to 1.
.LP
.KS
Genexpr() knows that for an addition operation to produce an eight bit result,
only an eight bit add needs to be done.  The same goes for subtract and
multiply, but not for divide, for example.  It thus calls genexpr() twice,
each time with maxsize set to 1 but with rp set to 0.  The next thing
encountered on the way down is the \f(CWextend\fP nodes.  These are processed by
a function defined as shown:
.DS
\f(CWint doextend(node,sign,sizenow,rp,maxsize)
int sign,sizenow,maxsize;
struct treenode *node;
struct es_entry *rp;
.DE
.KE
.KS
.LP
The function takes the \f(CWnode\fP, \f(CWrp\fP, and \f(CWmaxsize\fP
arguments already discussed,
plus the \f(CWsign\fP and \f(CWsizenow\fP arguments.  
\f(CWsign\fP simply specifies whether
the extension is to be by sign extension or by zero filling, and \f(CWsizenow\fP
specifies how big the value already is.  So now the extend() function simply
compares \f(CWsizenow\fP with \f(CWmaxsize\fP,
and if no extension is really required
just calls genexpr() again.  If an extension is required it only extends as
far as necessary, for example, to 16 bits rather than 32.  And finally, it
can deposit results directly using the caller's mechanism.  Thus in the
previous example, neither of the \f(CWextend\fP nodes cause any code at all to
be generated, and the whole statement \f(CWa\ =\ b\ +\ c\fP translates to this:
.DS
\f(CWADDB3	_b,_c,_a
.DE
.KE
.KS
.LP
Consider another example of ``push signed byte quantity a on the stack''.
doextend() has enough information to generate just this instruction:
.DS
\f(CWCVTBL	_a,-(R14)
.DE
.KE
.KS
.LP
It can be seen that the \f(CWrp\fP mechanism has allowed integrating the 
\f(CWpush\fP
operation with the ``extend'' operation.  Finally, consider the example ``copy
unsigned byte quantity a to word quantity b''  doextend() generates this code
for this:
.DS
\f(CWMOVZBW	_a,_b
.DE
.KE
.NH 2
Register Allocation and the Stack
.LP
The VAX has only a limited number of registers to work with, and occasionally
there arise situations where not all temporary values can be held in registers
at once.  The first such situation is when the ``depth'' of
an expression becomes
too great, i.e. there are more temporary values than available registers.  The
second situation is when a function is called.  Functions can be called from
various contexts, and so they do not know what state of the register set is
in and simply overwrite registers at will (this is the method used by the
system compiler with which tcc is object module compatible).
.KS
.LP
Therefore, a method is needed to keep temporary values somewhere other than
in registers.  This is provided by the 
\f(CW(R14)+\fP addressing mode, which pops a
value off the stack, and the \f(CW@(R14)+\fP
addressing mode, which pops an address
off the stack and accesses the value at that address.  When a register
is needed and none is available, the register allocator will traverse the
expression stack from the bottom, and turn the first addressing mode which it
finds which uses a temporary register into a stack one, as shown in the
following table:
.sp
.TS
center;
c c c
l l l.
Old Mode	Code generated	New Mode
_
\f(CWRx	PUSHL Rx,     	(R14)+
(Rx)	PUSHL Rx,     	@(R14)+
y(Rx)	PUSHAB y(Rx)	@(R14)+
@y(Rx)	PUSHL y(Rx)	@(R14)+\fP
_
.TE
.KE
.sp
.LP
Since the lowest entry in the expression stack which can surrender a register
is thus transformed, the flushed values end up on the system stack in just the
order in which they will be needed, and so the ``pop'' addressing modes can be
used just like the original ones.
.LP
Some precautions are in order here.  First, at no point may two operands in the
same instruction use a ``pop'' addressing mode unless it is absolutely certain
that the VAX CPU will ``pop'' the values in the proper order.  Second, if the
value to be popped is less than a longword, then building the ``pop'' right into
an instruction may cause only part of the pushed longword to be popped.  Both
these situations are prevented by careful coding of the code generator, with
explicit popping of stack contents back into registers before use
where unavoidable.
.KS
.NH 2
Expression Translation Example
.LP
Having discussed the main aspects of the expression translator's operation,
we will illustrate how the concepts work together with a simple expression
translation example.  The expression to be translated (in context) is:
.DS
\f(CWshort a;
int *b;
int c[20];

main()
{
    a = *(b+1) + func(a) + c[5];
}
.DE
.KE
.KS
The actual tcc output for the expression follows:
.DS
\f(CWaddl3   $4,_b,r0
pushl   r0
cvtwl   _a,-(r14)
calls   $1,_func
addw2   *(r14)+,r0
addw3   _c+20,r0,_a
.DE
.KE
.LP
The first line computes the address of 
\f(CW*(b+1)\fP
and loads it into a register.
The second line flushes this register to the stack since calling a function
will overwrite it.  The next line extends a to a longword and pushes it on
the stack, then the function is called.  Finally, the two add operations are
performed using the addressing modes which have accumulated on the expression
stack: 
\f(CW*(r14)+\fP from the \f(CW*(b+1)\fP, and \f(CW_c+20\fP
from the \f(CWc[5]\fP.  The result
is immediately dumped in the destination location by the last instruction thanks
to the \f(CWrp\fP mechanism, and the additions are performed at word precision
thanks to the \f(CWmaxsize\fP
mechanism.  Note that the code isn't perfect:  The
code generator could have rearranged the expression to eliminate at least
the \f(CWpushl r0\fP
instruction, but this compiler is not optimizing and the local
optimizations it does make can only go so far.
.NH
Boolean Expression Code Generation
.LP
The C standard specifies that the evaluation of boolean expressions
must be ``short-circuited'' that is, evaluation of the expression must
proceed from left to right and stop as soon as the final result is
known.  This section explains how this is done in tcc.
.KS
.NH 2
The gen_logicexpr() Function
.LP
The function is the second of the three most important
functions in the code generator (the third is described in the next
section).  It is declared as follows:
.DS
\f(CWgen_logicexpr(node,ltrue,lfalse,lnext)
struct treenode *node;
int ltrue,lfalse,lnext;
.DE
.KE
.LP
The function can be called in two ways.  From the statement level, it
is called with three label arguments, representing the label to jump
to if the expression is true, the one to jump to if the expression is
false, and the number of any label which will be generated immediately
after the function returns (or zero if none).  From the expression
level, it is called with the three label numbers all set to zero, in
which case the generated code will return success or failure as the
value 1 or 0, and gen_logicexpr() will leave information on how to get
this value on the expression stack.
.KS
.NH 2
Handling of &&, |\h'.2m'|\h'.2m', and ! Operators
.LP
The handling of the boolean operators in gen_logicexpr() is extremely
simple.  The function simply generates labels and calls itself, as
appropriate.  For example, with ltrue, lfalse, and lnext all set as
shown in the previous section, the handling of the \f(CW!\fP operator is
simply this:
.DS
\f(CWgen_logicexpr(node->link1,lfalse,ltrue,lnext);
.DE
.KE
.KS
.LP
The true and false labels have simply been reversed, reversing the
result of whatever the argument of 
\f(CW!\fP returns.  The handling of the \f(CW||\fP
operator is only slightly more complex:
.DS
\f(CWlab = labelcnt++;
gen_logicexpr(node->link1,ltrue,lab,lab);
genlabel(lab);
gen_logicexpr(node->link2,ltrue,lfalse,lnext);
.DE
.KE
.LP
Here, the first line generates a new label number.  The next line
processes the left argument of the \f(CW||\fP, with label numbers set so
that a false return causes the next expression to be evaluated, and a
true return causes the entire expression to become true.  The next
line generates the label, which is duly indicated in the \f(CWlnext\fP field
of the gen_logicexpr() call.  Finally, the right side of the \f(CW||\fP
operator is evaluated, with the label fields set so that the entire
expression succeeds or fails depending on the result.  Since we
are about to return, lnext can be set to whatever was given for the
whole expression.  It should be clear at this point how the \f(CW&&\fP
operator would work, and the code for it is not shown.
.NH 2
Handling of Comparison Operators
.LP
The six comparison operators are all conceptually handled in the same
way.  Both arguments are evaluated and deposited on the expression
stack by calls to genexpr(), then the appropriate comparison instruction
is generated, and the arguments are popped off the stack.
.LP
The first detail to point out is the method used to choose the size of
the comparison necessary.  This involves looking down both subtrees,
past any \f(CWextend\fP nodes resulting from the inevitable integer
promotion of the arguments, to see how big the arguments really are.
Then a compare instruction of the size of the bigger operand is
chosen, and genexpr() is called with \f(CWmaxsize\fP set to that size.
Also, when determining whether comparison instructions should be
signed or unsigned, the signed/unsigned property of the larger of the
two operands is used, unless they are equal in size in which case the
comparison is unsigned if either of the two is unsigned.
.LP
For example, if the unsigned short integer \f(CWa\fP and the signed character
\f(CWb\fP are to be compared, then \f(CWmaxsize\fP
is 2 and the comparison will be unsigned.
.KS
.NH 2
Efficient Generation of Conditional Branches
.LP
Nominally, each conditional branch generated in a logic expression
would take this form:
.DS
\(bu Evaluate condition
\(bu Branch to ltrue if true
\(bu Branch to lfalse if false
.DE
.LP
However, it is often the case that either ltrue or lfalse immediately
follow the above, and thus one of the branches may be unnecessary.  It
is clear that to reliably eliminate the second branch in this case,
the code generator must have access to both a ``branch if true'' and a
``branch if false'' version of each branch.  Then, if ltrue immediately
follows, this code can be generated:
.DS
\(bu Evaluate condition
\(bu Branch to lfalse if false
.DE
.KE
.KS
.LP
And if lfalse immediately follows, this code can be generated:
.DS
\(bu Evaluate condition
\(bu Branch to ltrue if true
.DE
.KE
.KS
.LP
Finally, if neither ltrue nor lfalse follow immediately, then the code
shown earlier must be generated, except that the second branch can be
unconditional.  All this work is done by the gencjmp() function, which
is defined as follows:
.DS
\f(CWgencjmp(optrue,opfalse,ltrue,lfalse,lnext)
char *optrue,*opfalse;
int ltrue,lfalse,lnext;
.DE
.KE
.KS
.LP
Its arguments are, respectively, the branch instruction which will
branch if the condition is true, the one which will branch if the
condition is false, and the three label numbers discussed earlier.
For example, the following call to gencjmp() generates code for ``jump
to l1 if the zero flag is set, l2 if it is not set, with l3 following
next'':
.DS
\f(CWgencjmp("jeql","jneq",l1,l2,l3);
.DE
The function will generate one of the following forms of code,
depending on the relationship of l1, l2, and l3:
.DS
\f(CW        jeql    l1          jneq    l2          jeql    l1
    l2: ...             l1: ...                 jbr     l2
                                                ...       
.DE
.KE
.NH 2
Returning Boolean Values
.LP
When gen_logicexpr() is first invoked, it checks if ltrue, lfalse, and
lnext are provided (i.e. non-zero).  When this is the case, it has
been called either from itself, or from the statement translator, and
need only transfer control to labels already provided.  But if they
are zero, it has been called from the expression translator, which
wants the result given as a value on the expression stack.
.LP
In this case, gen_logicexpr() allocates its own labels at the start,
and when the expression is finished, generates the labels along with
code to load 0 or 1 into a register.  Then it pushes that register on
the expression stack and thus returns a value to the caller.  The \f(CWrp\fP
and \f(CWmaxsize\fP mechanism is not provided here, so that the value is
always returned in a register.
.KS
.NH 2
Boolean Expression Example
.LP
For an example of boolean expression translation, we will consider the
following code:
.DS
\f(CWchar a,b;
unsigned short c;
int d,e;

main()
{
    e = !(a<c) || b && (c!=d);
}
.DE
The actual tcc output for the expression above is shown below:
.DS
\f(CW        cvtbw   _a,r0
        cmpw    r0,_c
        jgequ   L1
L3:     tstb    _b
        jeql    L2
L4:     movzwl  _c,r0
        cmpl    r0,_d
        jeql    L2
L1:     movl    $1,r0
        brb     L5
L2:     clrl    r0
L5:     movl    r0,_e
.DE
.KE
.LP
First, \f(CWa<c\fP is processed.  Since 
\f(CWc\fP is longer than \f(CWa\fP, \f(CWa\fP is extended to the size of 
\f(CWc\fP, and since \f(CWc\fP
is unsigned, an unsigned branch is used.  Due to the 
\f(CW!\fP and \f(CW||\fP
operators, the expression immediately succeeds if 
\f(CWa >= c\fP, and the conditional branch generated reflects exactly this,
branching to the point where 1 is returned from the expression.  Next,
\f(CWb\fP
is tested, and if it is zero, then the expression immediately fails,
returning zero.  Finally, 
\f(CWc\fP and \f(CWd\fP
are brought to a common size and
compared, and if they are equal, then the expression again fails.
Finally, control falls through to the point where 1 is returned.  It
is seen how the problem of branches branching around other branches
has been entirely avoided.
.NH
Statement Code Generation
.LP
The previous section explained the most difficult aspect of the code
generator, namely, how translation works at the expression level.  This
section describes how translation works at the statement level.
.KS
.NH 2
The codegen() Function
.LP
At the highest level, the code generator is visible as a single
function (the most important one), which is defined as follows:
.DS
\f(CWcodegen(node,lab1,lab2)
struct treenode *node;
int lab1,lab2;
.DE
.KE
.KS
.LP
The function takes as its argument a tree node which must be the root
of a parse tree representing a complete statement.  The other two
arguments are label numbers, indicating respectively the place where a
\f(CWbreak\fP statement and a \f(CWcontinue\fP
statement should transfer control.
These arguments are zero when the respective statement is not valid,
i.e. when there is no enclosing statement providing a place for it
to jump to.  The parser initially calls codegen like this:
.DS
\f(CWcodegen(node,0,0);
.DE
Where \f(CWnode\fP
is the root of the entire compound statement making up the body of a function.
.KE
.LP
The codegen() function handles flow-of-control constructs, compound
statements, and individual ``expression statements''  For example, for
a compound statement, it traverses the linked list of statements
inside, calling itself for each.  For an expression statement, it
calls genexpr(), then discards the result genexpr() leaves on the
expression stack by simply popping it off.  Actually, it calls a
special front end to genexpr() when the result is not required.  This
front end, called eval_and_toss(), handles certain expressions slightly
better, knowing that their results will not be needed.
.KS
.NH 2
Control Flow Construct Compilation
.LP
The fact that the codegen() function is recursive, and that it has the
gen_logicexpr() function at its disposal to generate conditional
branches, makes the handling of control flow constructs very simple.
We will show as an example the entire part of codegen() responsible
for \f(CWwhile\fP loops:
.DS
\f(CWl1 = labelcnt++;
l2 = labelcnt++;
l3 = labelcnt++;
genlabel(l1);
gen_logicexpr(node->link2,l2,l3,l2);
genlabel(l2);
codegen(node->link3,l3,l1);
en.type = ES_STATIC;
en.u.val1 = l1;
en.val2 = 0;
geninst("jbr",1,&en);
genlabel(l3);
.DE
.KE
.LP
The first three lines simply reserve three label numbers, which are
respectively: top of loop, ``true'' label for the conditional
expression, and end of loop.  Then the first label is generated, and the
conditional expression is translated so that control transfers to the
next statement if it is true, and to the end of the loop if it is
false.  Then the ``true'' label is generated, and the body of the loop
is compiled via a recursive call to codegen().  Finally, a branch to
the top of the loop is generated (four lines), and the end label is
emitted.  It is seen how the proper label numbers are passed down in
the recursive call to codegen(), directing 
\f(CWbreak\fP and \f(CWcontinue\fP statements to their places.
.LP
The code to compile the other control flow constructs is equally
simple, and so it will not be specifically discussed.
.NH 2
Switch Statement Implementation
.LP
A great deal of care was taken to have an efficient implementation
of the C language \f(CWswitch\fP statement.  This statement is best
implemented as a dispatch table, with an entry corresponding to each
of the possible values of the selector operand, with all those entries
for which no case was provided branching to the default location.
Unfortunately, not all switch statements have case values which are
closely clustered together and permit this type of thing.  For this
type, tcc builds an efficient compare-and-branch dispatcher, which
will be described.
.LP
Compilation of switch statements begins by a traversal of the parse
tree below the switch node, extending down to the individual statement
level but not going into the compound statements attached to switch
nodes.  A label number is assigned to each case, and the case values
and label numbers are collected in a dynamically allocated array
(therefore, there is no limit on the number of cases).
Next, the maximum and minimum case value in the array are found, the
array is sorted by case value, and duplicates are eliminated.  Then it
is time to decide which dispatch implementation will be used.  The
criterion is to use the dispatch table if there are more than two
cases and if the dispatch table will have at least one entry out of 10
going to a case (with the others going to the default).  If both are
not met, then the compare-and-branch dispatcher is used.
.KS
.NH 3
Dispatch Table Implementation
.LP
The dispatch table implementation uses a VAX \f(CWcase\fP instruction.  The
instruction is not described here, but an example is shown for the
following code fragment:
.DS
\f(CWswitch(a) {
    case 1:  < statement at L1 >
    case 2:  < statement at L2 >
    case 5:  < statement at L3 >
    default: < statement at L4 >
}
.DE
The tcc output corresponding to the above skeleton would be:
.DS
\f(CW        subl2   $4,sp
        casel   _a,$1,$4
L5:     .word   L1-L5
        .word   L2-L5
        .word   L4-L5
        .word   L4-L5
        .word   L3-L5
        jbr     L4
.DE
It is assumed that the reader is familiar with the VAX \f(CWcase\fP
instruction and thus this will not be discussed further.
.KE
.KS
.NH 3
Compare-and-branch Implementation
.LP
When each case value has to be explicitly tested against, the dispatch
time for the switch statement becomes painfully dependent on the
number of different cases.  For a straightforward compare/branch on
each case value, the dispatch time would obviously be O(n).
Fortunately, there is a simple method, requiring only a little more
code, with which this dispatch time can be reduced to O(log\s-3\d2\u\s0(n)), and
this was used for tcc.  The method is to generate a binary tree-shaped
dispatch system, in which the test against each case value eliminates
half the remaining case values.  A three-way branch is required after
each compare:  either go to the code for that case, or go check
lower values, or go check higher values.  The routine which does this
is a straightforward binary search of the case array, except that
instead of searching it generates code which traces its steps.  Its
output will be shown for the following skeleton:
.DS
\f(CWswitch(a) {
    case 1:         < statement at L1 >
    case 4:         < statement at L2 >
    case 16:        < statement at L3 >
    case 64:        < statement at L4 >
    case 256:       < statement at L5 >
    case 1024:      < statement at L6 >
    case 4096:      < statement at L7 >
    case 16384:     < statement at L8 >
    default:        < statement at L9 >
}
.DE
.KE
.KS
.LP
The tcc output for this skeleton would be as shown below (in two
columns to save space):
.DS
\f(CW        movl    _a,r0               jbr     L9
        cmpl    r0,$64      L11:    cmpl    r0,$1024
        jeql    L4                  jeql    L6
        jgtr    L11                 jgtr    L12
        cmpl    r0,$4               cmpl    r0,$256
        jeql    L2                  jeql    L5
        jgtr    L10                 jbr     L9
        cmpl    r0,$1       L12:    cmpl    r0,$4096
        jeql    L1                  jeql    L7
        jbr     L9                  cmpl    r0,$16384
L10:    cmpl    r0,$16              jeql    L8
        jeql    L3                  jbr     L9
.DE
.KE
.LP
As can be seen (by tracing through the code for various values of a),
no more than three comparisons are ever necessary to reach the proper
case or the default.
.NH
Code Generator Summary
.LP
There are many other functions in the code generator, but there is no
space in this document to describe them all.  It is hoped that the
description given in this section has given the reader the knowledge
necessary to understand the source code of the code generator by
describing major highlights throughout, and the philosophy used in the
design of the individual routines.  This is a hard-coded as opposed to
a table driven code generator primarily because it was thought that
the design for a table driven scheme, the implementation of the table
interpreters, and the writing of the tables would be far more time
consuming than writing the somewhat less than 2000 lines of C which
the code generator presently consists of has been.  Indeed, it was
possible to debug the code generator, as it was written, a small piece
at a time. Less than 30 lines of code had to be changed in further
debugging once the whole compiler got off the ground and large-scale tests
of more than 10-20 lines could finally be compiled.
