.CT
Compiler Limitations
.LP
This section describes limitations in the design, and encountered
during the construction of the compiler.
.NH 0
Design Limitations
.LP
The following points describe areas in the definition of the C language
that were not included in the design of the compiler.
.IP 1.
Floating point (data types float, double) is not supported by the compiler.
.IP 2.
Aggregate objects (structures, unions, arrays) cannot be passed to functions
as parameters or returned from functions.  All data types passed to functions
or returned from functions must fit in 32 bits (therefore, a long integer or
a pointer is the biggest object which can be passed).
.IP 3.
ANSI style function definitions are not supported.  ANSI style function
definitions have their entire parameter declarations in the parentheses of the
function declarator, and can be predeclared for argument type checking and
promotion.  The compiler only supports ``old style'' functions, in which
the function can be predeclared as to return value only, and in which all
arguments are extended to sizeof(int) and passed in equal sized slots on the
stack.
.IP 4.
Bit fields in structure definitions are not supported.
.IP 5.
The keywords \f(CWconst\fP and \f(CWvolatile\fP are not supported.  At present,
they are parsed by the declaration processor, and stored in the internal symbol
table data structures.  But the type checker/semantic analyzer, whose job it
would be to prevent writing to \f(CWconst\fP objects, does not do so at present.
The \f(CWvolatile\fP
keyword is unneeded at any rate, as the code generator does not
optimize sufficiently to make it necessary.
.IP 6.
Register variables are not supported.  Any \f(CWregister\fP
keywords are treated exactly like \f(CWauto\fP keywords.
.NH
Implementation Limitations
.LP
The following points describe limitations that came
into being during the construction of the compiler.
.KS
.IP 1.
Due to the design of the ANSI C grammar used for the parser, it is not
possible to fully support the scope rules of a user defined type.  This is
because the grammar depends on the lexical analyzer to distinguish between
type identifiers and regular identifiers by consulting the symbol table, and
return one of two tokens appropriately.  The grammar further lacks the
facilities to parse a declaration in which type identifiers are present as
anything but declaration specifiers.  This makes it impossible to ``hide'' a
defined type in an inner scope, as there is no way to process the declaration
which would hide it.  As an example, the following program will fail to
compile, though it is a valid ANSI C program.
.DS
\f(CWtypedef int INT;
INT a;
 
main()
{
   int INT = 1;
}
.DE
.KE
.KS
.IP 2.
The code generator does not support conditional expressions returning a
value which cannot be stored in a register.  A register will be allocated to
store the result, then both alternative expressions will be translated and
forced to each return their result in this common register.  A control
structure similar to that of an IF/ELSE is built to select between the
expressions.  Since structure/union types can not normally be held in a
register, and \f(CWvoid\fP
types have no value to hold in a register, the type
checker prevents such types from reaching the code generator.  The following
code fragments illustrate the problem constructs:
.DS
\f(CWstruct { ... } a,b,c;
void func();
int d,e;
\&...
a = d ? b : c;
d ? (void) e++ : func();
.DE
.KE
.KS
This does not impose an undue limitation on the programmer, however, since
the conditional expressions could easily be rewritten like this, for example:
.DS
\f(CWa = *(d ? &b : &c);
if(d) e++; else func();
.DE
.KE
.IP 3.
Unsigned long versions of the modulo (\f(CW%\fP) and right shift (\f(CW>>\fP)
operations are not supported.  This is because the VAX does not have unsigned
division or unsigned right shift operations in its instruction set.  The
existing C compilers perform unsigned long right shifting with a ''bit field
extract'' instruction, and unsigned long modulo using a library function to
do the unsigned division.  It was arbitrarily decided not to bother supporting
these in the code generator for tcc, especially as unsigned types were an
optional part of the project.  Unsigned division is fully supported, however,
including those cases where the library function must be called.
.IP 4.
For array dimensions, the internal representations of \f(CW[]\fP and \f(CW[0]\fP
are the same. It appears that the existing C compilers make a distinction here,
allowing arrays of size zero to be allocated as global variables or structure
members (no storage is reserved for them in either case), but not allowing the
same thing to be done with incompletely defined arrays.  tcc considers all
arrays of size zero to be incompletely defined, and does not allow such arrays
to be used in a context where their size must be known.
.KS
.IP 5.
The constant expression eliminator is guaranteed to eliminate expressions
which are entirely constant, but not guaranteed to spot and reduce all constant
subexpressions.  This is due to the order in which the parser recognizes
the expressions.  As an extreme case, the expression
.DS
\f(CWa + 1 + 2 + 3
.DE
is parsed in this order:
.DS
\f(CW(((a + 1) + 2) + 3)
.DE
due to the left-to-right associativity of the \f(CW+\fP
operator.  It thus becomes
non-constant right away and code is generated to perform all four additions.
It would take major work to improve on this because the commutativity of
\f(CW+\fP
and other operators would have to be taken into account and the parse tree
rearranged before things as the above example could be simplified.  In
the mean time, the programmer writing tcc can easily help the compiler
along by separating constant subexpressions in parentheses; then they
are always seen and reduced.  Thus the previous example would become
.DS
\f(CWa + (1 + 2 + 3)
.DE
.KE
.IP 6.
It is not possible to define an enumerated type and initialize a variable
to one of the enumerators on the same line.  This means that the following will
not work:
.DS
\f(CWenum a { one, two, three } b = one;
.DE
The reason for this is that an entire declaration is parsed by the parser and
then handed to the declaration processor as a unit.  Thus the left side of
the declaration has not been processed by the time the right side is parsed,
and thus "one" is not yet known as an enumerator.  Thus an error will be
generated for undefined identifier "one".  This could be fixed by splitting
out the processing of \f(CWstruct\fP, \f(CWunion\fP, and \f(CWenum\fP
type specifiers to be
performed immediately after such a type specifier has been parsed, then passing
a pointer to the resulting symbol table entry to the declaration processor.
It is estimated that perhaps 50 lines of code in the area of
\f(CWhandle_struct\fP, \f(CWhandle_enum\fP, and
\f(CWprocess_declaration_specs\fP would have to be changed to do this.
.IP 7.
The construct \f(CW"string1" "string2"\fP in ANSI C must be concatenated to form
\f(CW"string1string2"\fP.  Tcc does not do this.
It could be added by modifying the
string gathering function in the lexical analyzer to skip white space after it
sees the closing quote, and continue gathering if the first thing after the
white space is another quote.  It is estimated that perhaps 30 lines of code
might be required.  Tcc does support the old style convention of a
backslash/newline combination being ignored in a string, allowing very long
string constants to be built that way.
.KS
.IP 8.
The constant expressions for initializers are restricted.  Integers
may only be initialized to integer constant expressions.  Pointers may be
initialized to one of the following:
.DS
\(bu an integer constant expression cast to the pointer type
\(bu an integer constant expression evaluating to zero
\(bu the address of an identifier, taken with a single \f(CW&\fP
\(bu the name of a function or array (in which case the \f(CW&\fP is implied)
\(bu a quoted string (for character pointers)
.DE
.KE
.KS
These have been chosen as a best compromise between ease of translation and
usefulness.  They provide for dispatch tables to functions, and tables of
ASCII strings, for example.  But it is illegal to have expressions such as
the following (valid in ANSI C) in initializers compiled by tcc:
.DS
\f(CW&id1 - &id2 (to initialize an integer)
(int) &id1
&id1 + 1    (to initialize a pointer)
&array[5]
.DE
All other aspects of initialization, including initialization of character
arrays using quoted strings, and the rules for grouping of initializer
expressions using \f(CW{\fP and \f(CW}\fP, are fully supported.
.KE
.IP 9.
Enumeration types have been supported as a distinct type all the way through
type checking.  This makes it possible to issue a warning when a value whose
type is not the proper enumerated on is assigned to an enumerated variable.
Unfortunately, it necessitates many implicit conversions between
\f(CWenum\fP and \f(CWint\fP in the type checker, so that
\f(CWenum + enum\fP returns \f(CWint\fP, for example.
It is likely that this is not completely handled and that situations can be
found where integers work and enumerators will not.  At this writing, for
example, it is impossible to add an enumerated variable to an array
acting as a pointer.
.IP 10.
The declaration processor does not check for duplicate member names in 
a structure or union, duplicate enumerators in an enumeration, or a conflict
between an enumerator and some other identifier.  This is harmless:  two
structure members or enumerators with the same name cause the second one to
be invisible, and a conflict between an enumerator and another identifier always
causes that other identifier to be seen first.  But warnings could be generated
in such cases, with the addition of another 30 lines of code or so to the
declaration processor and a small run-time performance penalty.
.IP 11.
Fixed-size arrays are used in various places, and a couple of these are not
checked for overrun.  It is possible to crash the compiler with a string
constant longer than 1000 characters, or a file name longer than 80
characters.  These limitations can be increased by editing
``scan.l'' and ``main.c''
and recompiling.  An additional 20 lines of code or so added to the string
gathering function and the mainline would remove this potential problem.
.IP 12.
The code generator does not use the ``index'' mode of the VAX instruction set
at all.  Supporting this would have necessitated addressing modes which tie
up two registers (one for the base addressing mode and one for the index) and
this was judged too complex.  In addition, the index mode generates an offset
which is dependent on the size of the operation, which is not known at the
time the addressing mode is built, so structural changes to the code generator
would be necessary as well.
.KS
.IP 13.
It is not presently possible to pass options to the C preprocessor when it
is invoked automatically by tcc.  If such options are required, the preprocessor
must be invoked explicitly like this:
.DS
\f(CW/lib/cpp {options} file.c | tcc -n -o file.s {other options}
.DE
This is judged sufficient, since preprocessor options are rarely used outside
a Makefile, in which the syntax above would be acceptable.  In any case, a
20-30 line modification to the mainline would fix this problem.
.KE
.IP 14.
The error messages from the compiler are clumsy at best.  The error
messages generated at parse time attempt to point out the error
location in the source code; however, due to the order in which the
code is parsed the errors are only seen at the close of a syntactical
construct.  For example, in a \f(CWwhile\fP loop, where the argument to
\f(CWwhile\fP
has an invalid type, the error will not be reported until the
end of the loop body, at which point the \f(CWwhile\fP
grammar rule is
finally reduced and the argument type checked.  The result is a
pointer to a totally irrelevant position from one to many lines below
the actual error.  It is likely that eliminating attempts to point out
the error location for all but syntax errors would be a good approach,
using instead the line numbers built into the parse tree nodes like is
done in the later stages of the compile process.  A feature which
points out at which token the error occurred, such as found in the
system C compiler, would also be worthwhile.  And finally, the error
productions in the C grammar are still insufficient to recover cleanly
from syntax errors.  There are cases where a single syntax error
causes the compiler to spew out ten error messages or more.
