.CT
Operation of the Front End
.NH
Lexical analysis
.LP
The front end of tcc is a standard (f)lex generated scanner.  It is
invoked by the parser via the function yylex(), which returns a token
or zero.  Some tokens have further information associated with them.
These are the following:
.IP 1.
Tokens which indicate numbers (\f(CWINTEGER\fP, \f(CWUNSINT\fP,
\f(CWLONGINT\fP, etc.)
return the actual number in the global variable yylval.
.IP 2.
Tokens which indicate character strings (\f(CWSINGLECHAR\fP, \f(CWSTRING\fP)
return the length of the string in yylval and leave the string
itself accessible in the global variable stringbuf.  Escape
sequences are processed before the string is returned.  The
string can contain embedded nulls and is not null terminated.
.IP 3.
Tokens which indicate identifiers collect the identifier string,
null-terminate it, and make a privately allocated copy of it.
A pointer to this copy is returned in yylval.  The scanner
examines the symbol table to find if the identifier is a user
defined type, and returns \f(CWID\fP or \f(CWTYPE_ID\fP as the token
appropriately.
.LP
The scanner collects its input as it is read, so it always has a copy
of the current input line.  When an input line ends, it checks if the
error routine has flagged an error on that line, and if this is the
case, the line is printed with a pointer to the error position
underneath.
.KS
.NH
Parsing
.LP
The parser for tcc is produced by yacc/bison from a straightforward
ANSI C grammar.  The parser builds a parse tree using a variety of
parse tree construction functions, which are also the interface to the
type checker.  The simplest parse tree construction function is
defined as follows:
.DS
\f(CWstruct treenode *make_plain(type,l1,l2,l3)
int type;
struct treenode *l1, *l2, *l3;
.DE
.KE
.LP
It takes pointers to up to three nodes, which will become children of
the newly created node.  It returns the pointer to the newly created
node, which is set to have the given type.  If the node has less than
three children, null pointers must be provided for the extra child
node fields.  Some places in the grammar call other parse tree
construction functions, which additionally do type checking and other
special processing.
.KS
.LP
Each parse tree node contains the following information:
.RS
.IP \(bu
the type of the node.
.IP \(bu
the line number in the source code which the syntactic construct
represented by the node was on.
.IP \(bu
the data type of the object represented by the node; this is a
pointer to a ``type chain'' which will be discussed later, or null
if not appropriate (in the case of control statement nodes).
.IP \(bu
a ``format'' field identifying the format of the rest of the node.
.IP \(bu
a variety of format-dependent fields (see ``nodes.h'').
.RE
.KE
.LP
Error recovery in the grammar is done by a few strategically placed
``error'' productions.  These are invoked when the input to the parser
cannot be matched by any syntactic construct known to the it.  An
error is reported in these cases and parsing continues with incorrect
parse tree segments suitably disposed of.  Parse trees are built up to
the level of a declaration or a function body, then they are processed
by calls to the declaration processor or code generator, and disposed
of.
.NH
Error Reporting
.LP
Error reporting in tcc is done by calling one of the error functions.
These are err() and errl().  err() takes two integer arguments.  The
first is a flag indicating whether the call is for an error (1) or a
warning (0).  The second is a number used to index into the table of
error messages.  Further arguments are needed to fill in details in
some error messages; these are provided in the same manner as for
printf().  err() takes the error location to be the current point in
the input text, and instructs the scanner to print out the line with a
pointer to the error when it has finished reading it.  errl() takes an
additional argument, which is a tree node.  It takes the line number
from the tree node, and indicates that the error is ``near'' that line.
err() and errl() count errors (but not warnings) and if any errors
occur during a compile, the output file is deleted at the end.  This
does not work if the compiler is compiling to the standard output.
.NH
Symbol Table and Type Chains
.LP
The C language has a number of basic types (such as integer and char)
and several ``type constructors'' which can be applied to a basic type
to make it a complex one.  These are, for example, ``pointer to'',
\&``array of'', and ``structure containing''.  Tcc internally represents
types as chains of type nodes, called type chains.  All type chains
end with a basic type node.  Type chains can merge into other type
chains, but in this case, a ``type ref'' node is inserted just before
the merge point to store information specific to that type chain, such
as whether the object represented is constant or volatile.
.KS
.LP
The symbol table implementation is very simple.  The symbol table
entries are kept in a singly linked list, and each entry contains,
among other things, the name of the object it defines, a pointer to
the type chain for the object, and a block level.  The block level is
zero for external objects, one for objects inside a function, and
greater than one for objects in inner blocks of a function.  The
following functions provide the interface between the symbol table and
the rest of the compiler:
.DS
\f(CWvoid init_symtab()

struct symbol *make_sym(level,name,entrytype)
int level, entrytype;
char *name;

void delete_sym_level(level)
int level;

struct symbol *find_sym(name,entrytype,minlevel)
char *name;
int entrytype, minlevel;

struct symbol *find_next(ptr,name,entrytype,minlevel)
struct symbol *ptr;
char *name;
int entrytype, minlevel;
.DE
.KE
.LP
The first function initializes the symbol table to empty.  The second
function creates a symbol of the indicated level, name, and entry
type, and links it into the symbol table in the proper place, that is,
together with the other entries of the same block level.  The third
function deletes nodes in the symbol table until it comes to a node
whose level is lower than the indicated one.  The last two functions
search for entries, given a name, a minimum block level, and an entry
type.  Entry types are bit masks, so a number of different entry types
can be specified by logical OR of the entry types desired.  Finally,
the ``name'' argument can be left at null to indicate that the first
appropriate symbol table entry should be returned.
.KS
.LP
The following diagrams show a few examples of symbol table entries and
type chains.  The first one would be produced by the following declaration:
.DS
\f(CWint *a[];
.DE
.PS
.ps 8
boxwid=.375i
boxht=.375i
linewid=0.25i
lineht=0.25i
A:box "\f(CWa\fP"
arrow
box dashed "\f(CW[]\fP"
arrow
box dashed "\f(CW*\fP"
arrow
box dashed wid 0.5i "\f(CWsigned\fP" "\f(CW32-bit\fP"
arrow <- up from A.n
arrow down from A.s
.PE
.sp
.LP
Here we see the symbol table entry for \f(CWa\fP, with an \f(CWarray of\fP
and a
\f(CWpointer to\fP constructor applied to the basic type \f(CWsigned 32-bit\fP,
which is the tcc internal representation of \f(CWint\fP.
.KE
.KS
Next, the symbol
table entry for a user defined type, and a use of that type, is shown:
.DS
\f(CWtypedef int a;
a b;
.DE
.PS
.ps 8
boxwid=.375i
boxht=.375i
linewid=0.25i
lineht=0.25i
A:box wid 0.5i "\f(CWa\fP" "\f(CWtypedef\fP"
arrow
B:box dashed wid 0.5i "\f(CWsigned\fP" "\f(CW32-bit\fP"
arrow down from A.s
box "\f(CWb\fP"
arrow right from last box.e
box dashed wid 0.5i "\f(CWtyperef\fP"
spline -> from last box.e right 0.2i to B.s
arrow <- up from A.n
arrow down from A.s
.PE
.sp
.LP
Here we see a special symbol table entry storing the type definition,
and a second symbol table entry making use of the type definition via
a \f(CWtype ref\fP node.
.KE
.KS
Two examples of type chains for structure
definitions follow next; the first one shows how the members of a
structure are stored, the second one shows how a named structure is
stored in the symbol table and how other symbol table entries can use
it.  In this case, \f(CWe\fP does not require a \f(CWtype ref\fP
node since the structure was defined together with \f(CWe\fP.
.DS
\f(CWstruct { int a; char b; } c;
.DE
.PS
.ps 8
boxwid=.375i
boxht=.375i
linewid=0.25i
lineht=0.25i
A:box "\f(CWc\fP"
arrow
box dashed wid 0.5i "\f(CWstruct\fP" "\f(CWheader\fP"
arrow "members" down from last box.s
B:box dashed "\f(CWa\fP"
arrow right from last box.e
box dashed wid 0.5i "\f(CWsigned\fP" "\f(CW32-bit\fP"
move to B.s
arrow down
box dashed "\f(CWb\fP"
arrow right from last box.e
box dashed wid 0.5i "\f(CWsigned\fP" "\f(CW8-bit\fP"
arrow <- up from A.n
arrow down from A.s
.PE
.KE
.KS
.DS
\f(CWstruct d { ... } e;
.DE
.PS
.ps 8
boxwid=.375i
boxht=.375i
linewid=0.25i
lineht=0.25i
A:box wid 0.5i "\f(CWd\fP" "\f(CWstruct\fP"
arrow
B:box dashed wid 0.5i "\f(CWstruct\fP" "\f(CWheader\fP"
arrow 
box invis wid 0.5i "members" 
arrow down from A.s
E:box "\f(CWe\fP"
spline -> from last box.e right 0.2i to B.s
arrow <- up from A.n
arrow down from E.s
.PE
.KE
.sp
.KS
.LP
Next, here is an example of a declaration with more than one
declarator; it shows how the type chains share a common end node.
Again, there are no \f(CWtype ref\fP nodes since the information which would
normally be stored in such a node (const and volatile flags) is common
to all the declared objects.
.DS
\f(CWint a, *b, **c;
.DE
.PS
.ps 8
boxwid=.375i
boxht=.375i
linewid=0.25i
lineht=0.25i
A:box "\f(CWa\fP"
arrow
REF:box dashed wid 0.5i "\f(CWsigned\fP" "\f(CW32-bit\fP"
arrow down from A.s
B:box "\f(CWb\fP"
arrow right from B.e
box dashed "\f(CW*\fP"
spline -> from last box.e right 0.2i to REF.s
arrow down from B.s
C:box "\f(CWc\fP"
arrow right from C.e
box dashed "\f(CW*\fP"
arrow right
box dashed "\f(CW*\fP"
spline -> from last box.e right 0.2i to REF.s
arrow down from C.s
arrow <- up from A.n
.PE
.KE
.sp
.KS
.LP
Lastly, the following diagram shows how an enumerated type is stored
internally; it has a symbol table entry
of its own so it can be found again in the future, similar to a
structure definition.
.DS
\f(CWenum a { b, c, d } e;
.DE
.PS
.ps 8
boxwid=.375i
boxht=.375i
linewid=0.25i
lineht=0.25i
A:box wid 0.5i "\f(CWa\fP" "\f(CWenum\fP"
arrow
B:box dashed wid 0.5i "\f(CWenum\fP" "\f(CWheader\fP"
arrow
box dashed "\f(CWb = 0\fP"
arrow
box dashed "\f(CWc = 1\fP"
arrow
box dashed "\f(CWd = 2\fP"
arrow down from A.s
E:box "\f(CWe\fP"
spline -> from last box.e right 0.2i to B.s
arrow <- up from A.n
arrow down from E.s
.PE
.KE
.sp
.LP
Type chains have a \f(CWreference count\fP in each node, which indicates how
many other nodes are linked to this node.  In the last example above,
the reference count of the \f(CWenum header\fP node would be 2, and the
reference count of the \f(CWb = 0\fP node would be 1.  Type chains are
deallocated by a function called freetype(), and this function stops
deallocating when it hits a node whose reference count is greater than 1,
decrementing its reference count instead.  When a new link to a
type node is created, its reference count must be incremented by one.
When a block level in the symbol table is deleted, a freetype() is
done on the types of all the entries deallocated, but usually, this
does not delete the type chains since by that time, the parse tree
will have many pointers to the chains and the reference counts will
have been incremented accordingly.
.KS
.LP
The symbol table and type chains are built by the declaration
processor, whose primary entry point is defined as follows:
.DS
\f(CWprocess_declaration(node)
struct treenode *node;
.DE
.LP
This is called by the parser whenever it has assembled the parse tree
for an entire declaration.  The single argument is just the root node
of the tree.  The declaration processor builds symbol table entries
and type chains for the declaration, and after it returns, the parse
tree may be deallocated.
.KE
.KS
.NH
Type Checking
.LP
Type checking is the act of ensuring that C operators are applied only
to types for which they are permitted, of ensuring that arguments to
operators are promoted as required to types the operator can handle,
and of determining what the type returned by the operator is.  In tcc,
typechecking is done at the same time as parse tree construction.  Each
parse tree node which corresponds to an expression which can return a
value is linked to a type chain (typically all or part of the type chain
of some symbol table entry) so that the type of the value can be
instantly determined.  Functions which build parse tree nodes
according to this method are briefly discussed here.  First, the functions
which make tree leaf nodes:
.DS
\f(CWstruct treenode *make_var(id)
struct treenode *id;

struct treenode *eval_sizeof(type,node)
struct treenode *node;

struct treenode *make_const(num,uflag,lflag)
long num;
int uflag,lflag;

struct treenode *make_strconst(loc,len)
char *loc;
int len;
.DE
.KE
.LP
These functions respectively make the leaf nodes corresponding to
identifiers, sizeof() operators (which always reduce to constant
nodes of integer type), integer constant nodes (of unsigned and/or
long type, as determined by the flags), and string constants.  Each of
these nodes is constructed with a type chain attached.  For the
make_var() function, if the identifier is a variable, the type chain
is that of the variable.  For the make_strconst() function, the type
chain is 'pointer to signed 8-bit'.  For the others, it is a basic
type node.  Next come the functions which construct non-leaf nodes,
and do the actual type checking.
.KS
.LP
The make_cond() takes care of
conditional operations, that is, operations which require or produce a
\&``true/false'' type of result.  These include the C constructs \f(CWfor\fP,
\f(CWwhile\fP, \f(CWif\fP, the conditional expression \f(CW?:\fP,
and the boolean
operators \f(CW&&\fP, \f(CW||\fP, and \f(CW!\fP.
Which of these operators is being
handled is passed in the ``type'' argument.  The function will construct
the appropriate node and attach the appropriate type chain to it, but
only after checking argument types and printing any necessary error
messages or warnings.
.DS
\f(CWstruct treenode *make_cond(type,l1,l2,l3)
int type;
struct treenode *l1, *l2, *l3;
.DE
.KE
.KS
The make_arg() is used to ensure that a function argument has the
appropriate type.  At present, this involves only checking that the
argument can be represented in a 32-bit machine word, and that it is
not of structure, union, or array type.
.DS
\f(CWstruct treenode *make_arg(type,l1,l2)
int type;
struct treenode *l1,*l2;
.DE
.KE
.KS
.LP
The make_incdec() is used to type check the \f(CW++\fP and
\f(CW--\fP operators.
It ensures that the operand is one to which the operation can legally
be applied.  If this is not the case, it deletes the operation and
returns its argument node.
.DS
\f(CWstruct treenode *make_incdec(type,l1)
int type;
struct treenode *l1;
.DE
.KE
.KS
.LP
The next function makes a function call node.  This involves making
sure the first argument is indeed of type ``function'', or of ``pointer
to function'' in which case it inserts a node to dereference the
pointer automatically.  Also, the task of this function is to
automatically declare any undefined functions as external and
returning integer.
.DS
\f(CWstruct treenode *make_functioncall(node1,node2)
struct treenode *node1,*node2;
.DE
.KE
.KS
.LP
Finally, a remaining function takes care of most of the other
operators in the C language.  This function is defined as follows:
.DS
\f(CWstruct treenode *make_inner(type,l1,l2,l3)
int type;
struct treenode *l1,*l2,*l3;
.DE
The basic task of this function, as of the others, is to generate a
tree node, which it does by a call to make_plain().  But then it
examines the operator for which the node is, and checks the types of
arguments before proceeding.
.KE
.KS
The basic method is to produce a ``cross
product'' of the argument types, and to handle each combination
separately.  To do this, a CROSS() macro has been defined, which takes
two arguments and combines them into a unique result.  Using this,
type checking proceeds according to the following example, which might
be for an ``add'' operation:
.DS
\f(CWswitch( CROSS(t1->type,t2->type,8) ) {
    case CROSS(TT_BASICTYPE,TT_ARRAYOF,8):
\fI
		... convert array to pointer (by taking its address), then
		proceed with the following:
\f(CW
    case CROSS(TT_BASICTYPE,TT_POINTERTO,8):
\fI
		... handle addition of an integer type to a pointer,
		generating a special ``add'' node which returns a pointer type.
\f(CW
    case CROSS(TT_ARRAYOF,TT_BASICTYPE,8):
\fI
		... convert array to pointer (by taking its address), then
		proceed with the following:
\f(CW
    case CROSS(TT_POINTERTO,TT_BASICTYPE,8):
\fI
		... handle addition of a pointer type to an integer as above.
\f(CW
    case CROSS(TT_BASICTYPE,TT_BASICTYPE,8):
\fI
		... handle normal integer addition, generating an ``add'' node
		which returns an integer type.
\f(CW}
.DE
.KE
.KS
We conclude this section by illustrating what a complete parse tree,
with type checking done, might conceptually look like.  Suppose the
following declarations were in effect:
.DS
\f(CWint a[10];
char *ptr;
int x;
.DE
.KE
.KS
Where a is a global variable, and ptr and x are local variables on the
stack.  Then suppose the following code fragment were processed:
.DS
\f(CWwhile(*ptr++) a[x] = a[x] + 1;
.DE
.KE
.KS
The parse tree for this is illustrated in the following picture.  Note
that this parse tree might be one statement in the compound statement
of a function, and could be translated directly to code by a call to
codegen() on its root, as discussed later. (in the diagram only the
typenodes, shown as dashed boxes, for the conditional expression are shown)
.PS
.ps 8
boxwid=.6i
boxht=.375i
linewid=0.25i
lineht=0.25i
diag_x=0.5i
diag_y=0.6i
A:box "\f(CWwhile\fP"
arrow <- up from A.n
B:box "\f(CWextend\fP" at A + (-diag_x,-diag_y)
C:box "\f(CWexpr\fP" "\f(CWstat\fP" at A + (diag_x,-diag_y)

D:box "\f(CW*\fP" at B + (-diag_x,-diag_y)
E:box "\f(CW++\fP" at D + (-diag_x,-diag_y)
F:box "\f(CWptr\fP" "\f(CW-8(fp)\fP" at E + (0,-diag_y)
arrow left from F.w
Type1: box dashed wid boxht "\f(CW*\fP"
arrow left from Type1.w
Type2: box dashed "\f(CWsigned\fP" "\f(CW8-bit\fP"
arrow from A.s to B.n
arrow from A.s to C.n
arrow from B.s to D.n
arrow from D.s to E.n
arrow from E.s to F.n
arrow left from B.w
Type3: box dashed "\f(CWsigned\fP" "\f(CW32-bit\fP"
spline -> from E.w left 0.2i to Type1.n
spline -> from D.w left 0.2i to Type2.n

G:box "\f(CW=\fP" at C + (diag_x,-diag_y)
H:box "\f(CW+\fP" at G + (diag_x*2,-diag_y)
arrow from G.s to H.n
arrow from C.s to G.n

I:box "\f(CW1\fP" "\f(CWconst\fP" at H + (diag_x,-diag_y)
arrow from H.s to I.n

Array1_1: box "\f(CW[]\fP" at G + (-diag_x*2,-diag_y)
Array1_2: box "\f(CWa\fP" "\f(CW_a\fP" at Array1_1 + (-diag_x,-diag_y)
Array1_3: box "\f(CWx\fP" "\f(CW-4(fp)\fP" at Array1_1 + (diag_x,-diag_y)
arrow from Array1_1.s to Array1_2.n
arrow from Array1_1.s to Array1_3.n

Array2_1: box "\f(CW[]\fP" at H + (-diag_x,-diag_y)
Array2_2: box "\f(CWa\fP" "\f(CW_a\fP" at Array2_1 + (-diag_x,-diag_y)
Array2_3: box "\f(CWx\fP" "\f(CW-4(fp)\fP" at Array2_1 + (diag_x,-diag_y)
arrow from Array2_1.s to Array2_2.n
arrow from Array2_1.s to Array2_3.n

arrow from H.s to Array2_1.n
arrow from G.s to Array1_1.n
.PE
.KE
.LP
The type chain for \f(CWptr\fP was taken from the symbol table
and is shown in dashed boxes at the lower left.  The \f(CW++\fP operator
accepts a pointer type, and returns the same type; therefore, the
\f(CWtype\fP field of its node points to the same type chain.  The \f(CW*\fP
operator accepts a pointer and returns the object to which the pointer
points, in this case, the \f(CWsigned 8-bit\fP node.  Finally, an \f(CWinteger
promotion\fP is necessary before proceeding further, and as a result,
the type checker has inserted an \f(CWextend\fP node to make the result a
32-bit type.  This is presented as the conditional argument to the
\f(CWwhile\fP node.  Note that the code generator will ignore this particular
\f(CWextend\fP node since, working its way down the tree rather than up, it
realizes that it can just as well test a byte quantity directly,
rather than extending to integer size and then performing an integer
operation.
.NH
Constant Expression Evaluation
.LP
Constant expression evaluation is a necessary part of any compiler,
since the values of certain expressions must be computable at compile
time.  In tcc, constant expression evaluation has been put into the
functions which create parse tree nodes and do type checking for
arithmetic operations (these are part of the make_inner() function).
These functions call another function named eval_const() early on.
eval_const() examines the operation and the arguments, and if the
arguments are all integer constant nodes (resulting from make_const()
functions, eval_sizeof() functions, and other eliminated constant
expressions), then it immediately evaluates the result and returns a
constant node.  When the type checking/node construction function sees
this, it simply returns the constant node.  Thus, at code generation
time, to check for and evaluate a constant expression, one need only
examine its tree node.  If it is not an integer constant node, then
something in the expression could not be evaluated and it was not
constant.
.LP
A side benefit of the above scheme is that it automatically eliminates
some constant subexpressions in nonconstant expressions, without any
additional effort required.
