#include "def.h"
#include "nodes.h"
#include "nodetypes.h"
#include "error.h"
#include "types.h"

static char *sccsid = "@(#) const.c 1.28@(#)";

/*
 *  This funtion "evaluates" a constant subexpression.  The real work is all
 *  done when the parse tree is constructed; this function just verifies that
 *  the expression was reduced to a constant.  It complains if the type
 *  is not integral, such as when it has been cast to a pointer as part of
 *  the constant expression.
 */
long get_const(node)
struct treenode *node;
{
	struct type *t1;

	if(node->format != NF_LEAF_CONST) {
		errl(NOT_INT_CONSTANT,node);
		return 0;
	} else {
		t1 = skiprefnodes(node->vtype);
		if((t1->type != TT_BASICTYPE || t1->tu.flags & TF_VOID) &&
		   (t1->type != TT_ENUMHEADER)) {
			errl(NOT_INT_CONSTANT,node);
			return 0;
		}
		return node->v_sval;
	}
}

/*
 *  The following evaluates a node of type TN_SIZEOF_EXP or TN_SIZEOF_TYPE.
 *  In the former case, it builds a type exprssion for the abstract type
 *  given, takes the size of it, and immediately scraps the type expression
 *  again.  In the latter case, it just takes the size of the expression.
 *
 *  The "type" argument is TN_SIZEOF_EXP or TN_SIZEOF_TYPE, which at this
 *  point only serves to choose which operation to do.
 *
 *  An "integer constant" node is constructed with the result in it.
 */ 
struct treenode *eval_sizeof(type,node)
struct treenode *node;
{
	struct type *t1, *t2;
	int sc;
	long siz;
	struct treenode *p1;

	if(type == TN_SIZEOF_TYPE) {
		/*
		 *  node -> TN_TYPENAME
		 *
		 *  Evaluate the size of this type, then kill the subtree and turn
		 *  the node into a constant node.
		 */
		t2 = process_declaration_specs(node->link1,&sc,0);
		p1 = node->link2;
		t1 = buildtypechain(&p1,t2,0);
		freetype(t2);
		siz = size(t1);
		freetype(t1);
		kill_tree(node);
	} else {
		/*
		 *  node -> expression with a type
		 */
		siz = size(node->vtype);
	}
	if(!siz) errl(WARN_SIZEOF_ZERO,node);
	return make_const(siz,0,0);
}

/*
 *  This function makes an "integer constant" tree node.  The arguments are
 *  these:
 *
 *   num:    The value of the constant
 *   uflag:  Whether the constant is unsigned
 *   lflag:  Whether the constant is long
 *
 *  If neither uflag nor lflag are set, then a type just big enough to contain
 *  the argument expression is chosen.
 */
struct treenode *make_const(num,uflag,lflag)
long num;
int uflag,lflag;
{
	long flags = 0;
	struct type *tnode = (struct type *) safealloc(sizeof(struct type));
	struct treenode *node=(struct treenode *)safealloc(sizeof(struct treenode));
	tnode->type = TT_BASICTYPE;
	tnode->refcount = 1;
	if(uflag) flags |= TF_UNSIGNED;
	if(lflag) flags |= TF_BIT32;
	else if(num >= -0x80 && num <= 0x7f) flags |= TF_BIT8;
	else if(num >= -0x8000 && num <= 0x7fff) flags |= TF_BIT16;
	else flags |= TF_BIT32;
	tnode->tu.flags = flags;
	node->format = NF_LEAF_CONST;
	node->v_uval = num;
	node->vtype = tnode;
	node->lineno = linecnt;
	node->ntype = TN_VALUE;
	return node;
}

/*
 *  This function makes a "string constant" tree node.  It makes its own copy
 *  of the string for that purpose.  The arguments are these:
 *
 *   loc:  Location of the string to copy
 *   len:  Length of the string (strlen() won't work since the string can have
 *         nulls in it and is not null terminated.
 */
struct treenode *make_strconst(loc,len)
char *loc;
int len;
{
	extern struct type *maketypenode();
	extern char *memcpy();
	struct type *tnode = (struct type *) safealloc(sizeof(struct type));
	
	struct treenode *node=(struct treenode *)safealloc(sizeof(struct treenode));
	node->ntype = TN_STRING;
	node->format = NF_LEAF_STRCONST;
	node->lineno = linecnt;

	node->v_strconst = memcpy(safealloc(len),loc,len);
	node->v_strlen = len;

	tnode->type = TT_BASICTYPE;
	tnode->tu.flags = TF_BIT8;
	tnode->refcount = 1;
	maketypenode(&tnode,TT_ARRAYOF);
	tnode->tu.array.numentries = len+1;
	tnode->tu.array.sizeeach = 1;
	node->vtype = tnode;
	return node;
}

/*
 *  This function makes a simpler, untyped treenode with a null terminated
 *  string in it.  This is used to store label names and such.
 */
struct treenode *make_str(type,string)
int type,string;
{
	struct treenode *node=(struct treenode *)safealloc(sizeof(struct treenode));
	node->ntype = type;
	node->format = NF_LEAF_STRING;
	node->lineno = linecnt;
	node->v_str = save_str(string);
	node->vtype = 0;
	return node;
}
