#include "def.h"
#include "nodes.h"
#include "symtab.h"
#include "types.h"
#include "nodetypes.h"
#include "error.h"

static char *sccsid = "@(#) function.c 1.30@(#)";

/*
 *  This routine handles a function header.  It takes these arguments:
 *
 *   header:  a tree node representing the function name, argument list, etc
 *            (which takes the form of a declaration, hopefully of
 *             function type)
 *	 decs:    a linked list of declaration nodes, representing the argument
 *            declaration.s
 *
 */
void proc_fnheader(header,decls)
struct treenode *header, *decls;
{
	struct symbol *sym, *sym2, *arglist;
	struct type *rtype, *t1, *t2;
	struct treenode *p1, *p2;
	int sc = TN_EXTERN;
	long siz, offset;
	/*
	 *  Process the declaration_specifiers.  If none present, assume 
	 *  "extern int".
	 */
	t1 = process_declaration_specs(header->link1,&sc,0);
	
	if(sc != TN_EXTERN && sc != TN_STATIC) {
		errl(STORCLASS,header);
		freetype(t1);
		return;
	}
	/*
	 *  Now process the declarator.  The grammar should already have enforced
     *  that there is a symbol in it.
	 */
	p1 = header->link2;

	rtype = buildtypechain(&p1,t1,&p2);
	freetype(t1);

	if(!p1) panic("symbol missing in function declarator");

	/*
	 *  The declarator had better be a function type, but the grammar didn't
	 *  enforce this so we must check here.
	 */
	if(rtype->type != TT_FUNCTIONRETURNING) {
		errl(ERR_SYNTAX,header->link2);
		freetype(rtype);
		return;
	}
	/*
     *  Status quo:
	 *    p1 -> TN_ID with the function name
     *    p2 -> TN_FUNCTIONDECL or TN_FNDECL_TYPELIST or TN_FNDECL_IDLIST
	 *    rtype -> "type" of this function, i.e. "function returning..."
	 *
     *  Find the function in the symbol table and check, if it is in there,
     *  that everything matches.
     */
    sym = find_sym(p1->v_str,ET_GLOBAL,0);

	if(!sym) {
		/*
		 *  This identifier has never been seen before, add it to the
		 *  symbol table as a function returning the appropriate type
		 *  and with the appropriate storage class.
		 */
		sym = make_sym(0,p1->v_str,ET_GLOBAL);
	} else {
		t1 = skiprefnodes(sym->type);
		/*
		 *  Identifier has been seen before.  If the previous version wasn't
		 *  a function, or has already been defined as a function, we have
	     *  a problem.
		 */
		if(t1->type != TT_FUNCTIONRETURNING ||
		   sym->storclass == TN_GLOBALDEFINED) {
			errl(REDECLARATION,p1,p1->v_str);
			freetype(rtype);
			return;
		} else {
			/*
			 *  Previous declaration was for a function, and the function
			 *  hasn't been defined yet.  See if the return types match.
			 *  If not, print a warning, change the return type, and continue.
			 */
			if(!typematch(rtype,sym->type)) {
				errl(WARN_FN_TYPE,p1,p1->v_str);
			}
			freetype(sym->type);
			/*
			 *  If storage class in predeclaration was "static", and the
			 *  storage class in this declaration isn't, print a warning.
			 */
			if(sc != TN_STATIC && sym->storclass == TN_STATIC) 
				errl(WARN_STATIC,p1,p1->v_str);
		}
	}
	/*
	 *  Now change the storage class in the symbol table to mark this
	 *  function as having been defined, with the appropriate return type.
	 */
	sym->type = rtype;
	sym->storclass = TN_GLOBALDEFINED;
	/*
	 *  Output the function prologue.
	 */
	fn_prologue(sym->name,(sc != TN_STATIC));
	/*
	 *  It remains to process the function argument list.
	 */
	switch(p2->ntype) {
		case TN_FUNCTIONDECL:
			/*
			 *  No args, old style.  Error if any declarations are given.
			 */
			if(decls) errl(ERR_SYNTAX,decls);
			break;
		case TN_FNDECL_TYPELIST:
			panic("ANSI style arg declarations not supported yet");
			break;
		case TN_FNDECL_IDLIST:
			/*
			 *  Gather up the identifiers in the function argument list and
			 *  string them together in the symbol table.
			 */
			sym2 = 0;
			p2 = p2->link2;
			while(p2) {
				sym = make_sym(1,p2->link2->v_str,ET_ARGUMENT);
				sym->type = 0;
				if(!sym2) arglist = sym; else sym2->su.auxlink = sym;
				sym2 = sym;
				p2 = p2->link1;
			}
			sym2->su.auxlink = 0;
			/*
			 *  argdecls is either null or points to a list of TN_DECLARATION.
			 */
			while(decls) {
				/*
				 *  Process the declaration specifiers
				 */
				sc = -1;
				t1 = process_declaration_specs(decls->link2,&sc,0);
				if(sc != -1 && sc != TN_REGISTER) errl(STORCLASS,decls);
				p1 = decls->link3;
				while(p1) {
					/*
					 *  Process one argument declarator.  First, build the
					 *  complete type expression.
					 */
					if(p1->link3) err(ERR_SYNTAX);
					p2 = p1->link2;
					t2 = skiprefnodes(buildtypechain(&p2,t1,0));
					/*
					 *  Transform function or array types into the appropriate
					 *  pointer type
					 */
					if(t2->type == TT_ARRAYOF) {
						struct type *t3 = t2->tn.next;
						t3->refcount++;
						maketypenode(&t3,TT_POINTERTO);
						freetype(t2);
						t2 = t3;
					} else if(t2->type == TT_FUNCTIONRETURNING)
						maketypenode(&t2,TT_POINTERTO);
					/*
					 *  Find the corresponding symbol table entry and attach
					 *  the type to it, checking for errors.
					 */
					if(!p2) panic("symbol missing in declarator");
					sym = find_sym(p2->v_str,ET_ARGUMENT,0);
					if(!sym) {
						errl(NOTFUNCARG,p2,p2->v_str);
						freetype(t2);
					} else if(sym->type) {
						errl(REDECLARATION,p2,p2->v_str);
						freetype(t2);
					} else sym->type = t2;
					p1 = p1->link1;
				}
				freetype(t1);
				decls = decls->link1;
			}
			/*
			 *  Now go through the argument nodes in order (by following the
			 *  chain collected earlier), make all which don't have a type yet
			 *  "int", and assign the offsets from the argument pointer,
			 *  starting at 4.
			 */
			offset = 4;
			sym = arglist;
			while(sym) {
				sym2 = sym->su.auxlink;
				if(!sym->type) {
					sym->type = (struct type *) safealloc(sizeof(struct type));
					sym->type->type = TT_BASICTYPE;
					sym->type->tu.flags = TF_BIT32;
					sym->type->refcount = 1;
				}
				sym->offset = offset;
				offset += 4;
				sym = sym2;
			}
			break;
		default:
			panic("bad function node");
	}
}
