#include "def.h"
#include "exprstack.h"
#include "nodetypes.h"
#include "nodes.h"
#include "symtab.h"
#include "error.h"
#include "types.h"

static char *sccsid = "@(#) genexpr.c 1.44@(#)";
extern int labelcnt;

/*
 *  These two variables describe the expression stack, which stores the
 *  addressing modes of instruction operands.  See "exprstack.h" for the
 *  format in which the modes are stored.
 */
struct es_entry expr_stack[ES_STACKDEPTH];
int es_pointer;
/*
 *  This is a bit map of available registers.
 */
long regmask;
/*
 *  This indicates whether a label has already been emitted on the current
 *  line.
 */
int labelwaiting;


/*
 *  This pushes an entry on the expression stack, checking for overflow.
 *  It returns the address of the entry.
 */
struct es_entry *es_push()
{
	if(++es_pointer >= ES_STACKDEPTH) panic("expr stack overflow");
	return(&expr_stack[es_pointer]);
}

/*
 *  This pops an entry off the expression stack, deallocating any registers
 *  the entry may be using (but not including stack pointer, frame pointer,
 *  arg pointer, etc.)
 */
es_pop(i)
int i;
{
	register struct es_entry *ep;
	while(i--) {
		if(es_pointer<0) panic("attempt to pop empty expr stack");
		ep = &expr_stack[es_pointer];
		if(ep->u.val1 < 6 && (ep->type == ES_REG
		   || ep->type == ES_REG+ES_DEFERRED || ep->type == ES_DREGOFFSET
		   || ep->type == ES_DREGOFFSET+ES_DEFERRED)) freereg(ep->u.val1);
		--es_pointer;
	}
}

/*
 *  This swaps the top two entries on the expression stack.
 */
es_swap()
{
	register struct es_entry en, *ep = &expr_stack[es_pointer-1];
	en = *ep;
	*ep = *(ep+1);
	*(ep+1) = en;
}

/*
 *  This function allocates a register.  First it looks in the map of available
 *  registers, and if it finds one, it clears the bit for that register and
 *  returns the register number.  Otherwise, it tries to free the least
 *  recently allocated register by flushing it onto the stack.  If that fails
 *  too, the compiler is broken.
 */
int getreg()
{
	register long m = 1;
	register int i;
	for(i=0;i<16;i++) {
		if(regmask & m) {
			regmask &= ~m;
			return i;
		} else
			m <<= 1;
	}
	i = flushregs(0);
	if(i<0) panic("out of registers");
	regmask &= ~(1<<i);
	return i;
}	

/*
 *  This returns a register to the map of free registers.
 */
freereg(i)
{
	long m = 1<<i;
	if(m & regmask) panic("attempt to free register not allocated");
	regmask |= m;
}

/*
 *  This emits an instruction of the form:
 *
 *   xxxxsn [arg1] [arg2] [arg3]
 *
 *  where xxxx is the opcode, s is the size (b, w, or l), and n is (optionally)
 *  the number of arguments.
 *
 *  n is not generated if the argument narg is zero.
 */
gensizop(base,size,narg,numargs,arg1,arg2,arg3)
char *base;
int size,narg,numargs,arg1,arg2,arg3;
{
	char buf[10];
	sprintf(buf,narg?"%s%c%d":"%s%c",base,"_bw_l"[size],narg);
	geninst(buf,numargs,arg1,arg2,arg3);
}

/*
 *  This routine emits an instruction.  It takes the following arguments:
 *
 *   code:		String representing the opcode
 *	 numargs:	Number of arguments provided.
 *   arg1-arg4: Arguments, given as pointers to expression stack entry nodes.
 */
geninst(code,numargs,arg1,arg2,arg3,arg4)
char *code;
int numargs;
struct es_entry *arg1, *arg2, *arg3, *arg4;
{
	int i;
	struct es_entry **p = &arg1;
	text();
	printf("\t%s",code);
	if(numargs) putchar('\t');
	for(i=1;i<=numargs;i++) {
		struct es_entry *ep = *p;
		switch(ep->type) {
			case ES_REG:
				printf("r%d",ep->u.val1);
				break;
			case ES_REG+ES_DEFERRED:
				printf("(r%d)",ep->u.val1);
				break;
			case ES_CONST:
				printf("$%ld",ep->u.val1);
				break;
			case ES_POSTINCR+ES_DEFERRED:
				putchar('*');
			case ES_POSTINCR:
				printf("(r%d)+",ep->u.val1);
				break;
			case ES_PREDECR:
				printf("-(r%d)",ep->u.val1);
				break;
			case ES_DREGOFFSET:
				if(ep->val2) printf("%ld(r%d)",ep->val2,ep->u.val1);
				else printf("(r%d)",ep->u.val1);
				break;
			case ES_DREGOFFSET+ES_DEFERRED:
				printf("*%ld(r%d)",ep->val2,ep->u.val1);
				break;
			case ES_LABELADDR:
				if(ep->val2) printf("$_%s+%ld",ep->u.str1,ep->val2);
				else printf("$_%s",ep->u.str1);
				break;
			case ES_LABEL+ES_DEFERRED:
				putchar('*');
			case ES_LABEL:
				if(ep->val2) printf("_%s+%ld",ep->u.str1,ep->val2);
				else printf("_%s",ep->u.str1);
				break;
			case ES_STATICADDR:
				if(ep->val2) printf("$L%d+%ld",ep->u.val1,ep->val2);
				else printf("$L%d",ep->u.val1);
				break;
			case ES_STATIC+ES_DEFERRED:
				putchar('*');
			case ES_STATIC:
				if(ep->val2) printf("L%d+%ld",ep->u.val1,ep->val2);
				else printf("L%d",ep->u.val1);
				break;
			default:
				panic("geninst: unknown addressing mode");
		}
		if(i<numargs) putchar(',');
		p++;
	}
	putchar('\n');
	labelwaiting = 0;
}

/*
 *  This function dereferences a pointer.  It converts an addressing mode to
 *  add a level of indirection, possibly generating an instruction if the
 *  addressing mode allows no further indirection.  The addressing mode in
 *  question is the top one on the expression stack.
 */
int indirect()
{
	register struct es_entry *ep = &expr_stack[es_pointer];
	struct es_entry en;

	switch(ep->type) {
		case ES_REG:
		case ES_POSTINCR:
		case ES_DREGOFFSET:
		case ES_LABEL:
		case ES_STATIC:
			ep->type |= ES_DEFERRED;
			break;
		case ES_LABELADDR:
			ep->type = ES_LABEL;
			break;
		case ES_STATICADDR:
			ep->type = ES_STATIC;
			break;
		case ES_DREGOFFSET+ES_DEFERRED:
		case ES_REG+ES_DEFERRED:
			if(ep->u.val1 < 6) {
				en.type = ES_REG;
				en.u.val1 = ep->u.val1;
				geninst("movl",2,ep,&en);
				ep->type = ES_REG+ES_DEFERRED;
				break;
			}
		case ES_CONST:
		case ES_POSTINCR+ES_DEFERRED:
		case ES_LABEL+ES_DEFERRED:
		case ES_STATIC+ES_DEFERRED:
			en.type = ES_REG;
			en.u.val1 = getreg();
			geninst("movl",2,ep,&en);
			ep->type = ES_REG+ES_DEFERRED;
			ep->u.val1 = en.u.val1;
			break;
		case ES_PREDECR:
			panic("indirect through predecr");
		default:
			panic("indirect: unknown addressing mode");
	}
}

/*
 *  This function is the opposite of "indirect" and takes the address of its
 *  operand.  The operand is on the top of the expression stack.  The function
 *  takes the usual rp and maxsize arguments; see "genexpr" for an explanation
 *  of these. It returns the usual success/fail condition as appropriate.
 */
addressof(rp,maxsize)
struct es_entry *rp;
{
	register struct es_entry *ep = &expr_stack[es_pointer];
	struct es_entry en;

	switch(ep->type) {
		case ES_REG:
		case ES_CONST:
		case ES_POSTINCR:
		case ES_PREDECR:
		case ES_STATICADDR:
		case ES_LABELADDR:
			panic("address of register or constant");
		case ES_REG+ES_DEFERRED:
		case ES_POSTINCR+ES_DEFERRED:
		case ES_DREGOFFSET+ES_DEFERRED:
		case ES_LABEL+ES_DEFERRED:
		case ES_STATIC+ES_DEFERRED:
			ep->type &= ~ES_DEFERRED;
			break;
		case ES_LABEL:
			ep->type = ES_LABELADDR;
			break;
		case ES_STATIC:
			ep->type = ES_STATICADDR;
			break;
		case ES_DREGOFFSET:
			if(rp && maxsize >= 4) {
				if(rp->type == ES_PREDECR && rp->u.val1 == REG_SP)
					geninst("pushab",1,ep);
				else
					geninst("movab",2,ep,rp);
				es_pop(1);
				return 1;
			} else {
				en.type = ES_REG;
				en.u.val1 = getreg();
				geninst("movab",2,ep,&en);
				es_pop(1);
				*es_push() = en;
			}
			break;
		default:
			panic("indirect: unknown addressing mode");
	}
	return 0;
}

/*
 *  This function adds an offset to the effective address of an opcode (which
 *  corresponds to the "." operator, or a constant subscript on an array, in
 *  C).  It generates code where this is not directly possible with the
 *  addressing mode.
 */
addoffset(offset)
long offset;
{
	struct es_entry en;
	register struct es_entry *ep = &expr_stack[es_pointer];
	if(!offset) return;
	switch(ep->type) {
		case ES_DREGOFFSET+ES_DEFERRED:
			if(ep->u.val1 < 6) {
				en.u.val1 = ep->u.val1;
				en.type = ES_REG;
				ep->type = ES_DREGOFFSET;
				geninst("movl",2,ep,&en);
				ep->val2 = offset;
				break;
			}
		case ES_POSTINCR+ES_DEFERRED:
		case ES_STATIC+ES_DEFERRED:
		case ES_LABEL+ES_DEFERRED:
			en.type = ES_REG;
			en.u.val1 = getreg();
			ep->type &= ~ES_DEFERRED;
			geninst("movl",2,ep,&en);
			ep->type = ES_DREGOFFSET;
			ep->u.val1 = en.u.val1;
			ep->val2 = offset;
			break;
		case ES_REG+ES_DEFERRED:
			ep->type = ES_DREGOFFSET;
			ep->val2 = offset;
			break;
		case ES_DREGOFFSET:
		case ES_LABEL:
		case ES_STATIC:
			ep->val2 += offset;
			break;
		default:
			printf("addr mode = %d\n",ep->type);
			panic("can't add offset to this addressing mode");
	}
}

/*
 *  This function flushes registers onto the stack.  Any addressing mode
 *  referring to a register is changed to refer to the stack instead, possibly
 *  simplifying it in the process.
 *
 *   all: 0 - flush just one register
 *        1 - flush all registers
 *
 *  -1 is returned unless just one register is flushed and this succeeds
 *  (this is used by the register allocator).
 */
int flushregs(all)
int all;
{
	register int i,r;
	struct es_entry en;
	register struct es_entry *ep;

	for(i=0;i<=es_pointer;i++) {
		ep = &expr_stack[i];
		if(ep->u.val1 < 6) {
			r = ep->u.val1;
			switch(ep->type) {
				case ES_REG:
					geninst("pushl",1,ep);
					ep->type = ES_POSTINCR;
					ep->u.val1 = REG_SP;
					break;
				case ES_REG+ES_DEFERRED:
					en.type = ES_REG;
					en.u.val1 = ep->u.val1;
					geninst("pushl",1,&en);
					ep->type = ES_POSTINCR+ES_DEFERRED;
					ep->u.val1 = REG_SP;
					break;
				case ES_DREGOFFSET:
					geninst("pushab",1,ep);
					ep->type = ES_POSTINCR+ES_DEFERRED;
					ep->u.val1 = REG_SP;
					break;
				case ES_DREGOFFSET+ES_DEFERRED:
					ep->type = ES_DREGOFFSET;
					geninst("pushl",1,ep);
					ep->type = ES_POSTINCR+ES_DEFERRED;
					ep->u.val1 = REG_SP;
					break;
				default:
					continue;
			}
			freereg(r);
			if(!all) return r;
		}
	}
	return -1;
}

/*
 *  This function handles "extend" nodes in the parse tree.  It takes the
 *  arguments 'node', which points to the tree node containing the extend,
 *  and 'sign' indicating whether the extension is a signed one.  It also
 *  takes the usual rp and sizemax arguments (see genexpr) and returns the
 *  usual true/false.
 *
 *  Extension nodes are avoided at all cost.  If the caller wants an object
 *  which is the same size as the unextended one, then nothing at all is
 *  done and genexpr is called again.  If it wants a smaller one, then 
 *  genexpr is called and the result loaded into a register.  If it wants
 *  a bigger one, then an actual extend instruction is generated.
 */
doextend(node,sign,sizenow,rp,sizemax)
int sign,sizenow,sizemax;
struct treenode *node;
struct es_entry *rp;
{
	struct es_entry en, *ep;
	char buf[10];
	if(sizemax == sizenow)
		return genexpr(node->link1,rp,sizenow);
	else
		genexpr(node->link1,0,sizenow);

	if(sizemax <= sizenow || (node->ntype == TN_VALUE
		&& node->format == NF_LEAF_CONST)) return 0;
	ep = &expr_stack[es_pointer];
	en = *ep;
	es_pop(1);
	sprintf(buf,sign?"cvt%c%c":"movz%c%c","_bw_l"[sizenow],"_bw_l"[sizemax]);
	if(!rp) {
		es_push();
		ep->type = ES_REG;
		ep->u.val1 = getreg();
		geninst(buf,2,&en,ep);
		return 0;
	} else {
		geninst(buf,2,&en,rp);
		return 1;
	}
}

/*
 *  This function determines whether an addressing mode contains a register
 *  which can be re-used.
 */
int isreg(ep)
struct es_entry *ep;
{
	int type = ep->type & ~ES_DEFERRED;
	if((type==ES_REG||type==ES_POSTINCR||type==ES_PREDECR||type==ES_DREGOFFSET)
		&& (ep->u.val1 < 6)) return 1;
	else return 0;
}

/*
 *  The following function generates code to return the complemented or
 *  negated result of an expression.  It is split out here because such a
 *  result is needed for several different operators.
 */
enum comneg_mode {COMPLEMENT,NEGATE};

int comneg(mode,node,rp,sizemax)
enum comneg_mode mode;
struct treenode *node;
struct es_entry *rp;
int sizemax;
{
	struct es_entry en, *ep;
	int r;
	char *root;
	genexpr(node,0,sizemax);
	ep = &expr_stack[es_pointer];
	if(ep->type == ES_CONST) {
		ep->u.val1 = (mode == COMPLEMENT) ? ~(ep->u.val1) : -(ep->u.val1);
		return 0;
	} else
		root = (mode == COMPLEMENT) ? "mcom" : "mneg";
		
	if(rp) {
		gensizop(root,sizemax,0,2,ep,rp);
		return 1;
	} else if(isreg(ep)) {
		en.type = ES_REG;
		en.u.val1 = ep->u.val1;
		gensizop(root,sizemax,0,2,ep,&en);
		*ep = en;
	} else {
		en.type = ES_REG;
		en.u.val1 = getreg();
		gensizop(root,sizemax,0,2,ep,&en);
		es_pop(1);
		*es_push() = en;
	}
	return 0;
}

/*
 *  This function generates code for a commutative binary operator (such
 *  as "add", "mul", "bis", etc.)  It takes the following arguments:
 *
 *   root:                the opcode without size or number of arg extensions
 *   leftnode, rightnode: the parse tree nodes representing the operands
 *   rp, maxsize:         the usual mechanism (see genexpr)
 */
int dobinop(root,leftnode,rightnode,rp,maxsize)
char *root;
struct treenode *leftnode,*rightnode;
int rp,maxsize;
{
	struct es_entry en,*ep;
	genexpr(leftnode,0,maxsize);
	genexpr(rightnode,0,maxsize);
	ep = &expr_stack[es_pointer-1];
	stackfix(ep,maxsize);
	if(rp) {
		if(samething(rp,ep)) gensizop(root,maxsize,2,2,ep+1,rp);
		else if(samething(rp,ep+1)) gensizop(root,maxsize,2,2,ep,rp);
		else gensizop(root,maxsize,3,3,ep+1,ep,rp);
		es_pop(2);
		return 1;
	} else if(ep->type == ES_REG) {
		gensizop(root,maxsize,2,2,ep+1,ep);
		es_pop(1);
	} else if((ep+1)->type == ES_REG) {
		gensizop(root,maxsize,2,2,ep,ep+1);
		*ep = *(ep+1);
		--es_pointer;
	} else {
		en.type = ES_REG;
		en.u.val1 = getreg();
		gensizop(root,maxsize,3,3,ep+1,ep,&en);
		es_pop(2);
		*es_push() = en;
	}
	return 0;
}

/*
 *  This function is like "dobinop" but for noncommutative operators (such
 *  as "sub", "div", "bic", etc).
 *
 *  It does not take left/right operand arguments since the caller must
 *  process these first and leave them on the expression stack.
 */
int dolrop(root,rp,maxsize)
char *root;
int rp,maxsize;
{
	struct es_entry en,*ep;
	ep = &expr_stack[es_pointer-1];
	if(rp) {
		if(samething(ep,rp)) gensizop(root,maxsize,2,2,ep+1,rp);
		else gensizop(root,maxsize,3,3,ep+1,ep,rp);
		es_pop(2);
		return 1;
	} else if(ep->type == ES_REG) {
		gensizop(root,maxsize,2,2,ep+1,ep);
		es_pop(1);
	} else if((ep+1)->type == ES_REG) {
		gensizop(root,maxsize,3,3,ep+1,ep,ep+1);
		es_pointer -= 1;
		es_pop(1);
		es_pointer += 1;
		*ep = *(ep+1);
	} else {
		en.type = ES_REG;
		en.u.val1 = getreg();
		gensizop(root,maxsize,3,3,ep+1,ep,&en);
		es_pop(2);
		*es_push() = en;
	}
	return 0;
}

/*
 *  This function takes a TN_STRING node in the parse tree and generates
 *  assembler pseudo-ops to form the string.  The arguments are as follows:
 *
 *   node:   the TN_STRING node
 *   length: the maximum number of characters to emit
 *   zflag:  whether the string should be null-terminated
 */
genstring(node,length,zflag)
struct treenode *node;
int length,zflag;
{
	register char c,*p = node->v_strconst;
	register enum {ASC,BYT,NUL} state;

	state = NUL;

	while(length-- >= !zflag) {
		c = (length<0) ? 0 : *p++;

		if(c>=' ' && c<127 && c != '"' ) switch(state) {
			case BYT: putchar('\n');
			case NUL: printf("\t.ascii\t\"");
					  state = ASC;
			case ASC: putchar(c);
		} else switch(state) {
			case ASC: printf("\"\n");
			case NUL: printf("\t.byte\t%d",c);
					  state = BYT;
					  break;
			case BYT: printf(",%d",c);
		}
	}
	if(state==ASC) printf("\"\n"); else putchar('\n');
}

/*
 *  This is the most central function in the code generator.  It generates code
 *  for an expression.
 *
 *  Nominally, the result of the expression would be left on the top of the
 *  expression stack.  However, often this would lead to a wasted instruction,
 *  such as in the case of "a+b", which would have no choice but to allocate
 *  a register, add a and b, and store the result in that register, even though
 *  the very next instruction to be generated would store this register
 *  somewhere where an "addl3" instruction could equally well have.
 *
 *  To prevent this, a mechanism has been implemented where all cooperating
 *  functions take the arguments "rp" and "maxsize", and return a boolean
 *  success flag.  This works as follows:
 *
 *  IF   the result can be generated in such a way that it can be stored
 *       in the destination "for free" (e.g. by turning an addl2 into an addl3)
 *
 *  AND  rp is not zero
 *  AND  the size of the result is maxsize
 *
 *  THEN store the result directly in the destination and return 1
 *  ELSE store the result into a register (or wherever) and return 0.
 *
 *  The only other argument 'node' is the parse tree node representing the
 *  expression.
 */
int genexpr(node,rp,maxsize)
register struct treenode *node;
register struct es_entry *rp;
{
	struct es_entry en, en1;
	register struct es_entry *ep;
	struct treenode *n1;
	int l1,l2,l3;
	int argcnt;
	int flag;
	int siz,siz1,lsiz;
	char ibuf[12];
	struct type *t1;

	if(!node) panic("codegen called with null pointer");

	switch(node->ntype) {
		case TN_SBYTETOLONG:
			return doextend(node,1,1,rp,maxsize);
		case TN_UBYTETOLONG:
			return doextend(node,0,1,rp,maxsize);
		case TN_SWORDTOLONG:
			return doextend(node,1,2,rp,maxsize);
		case TN_UWORDTOLONG:
			return doextend(node,0,2,rp,maxsize);
		case TN_STRING:
			data(1);
			printf("L%d:",labelcnt);
			genstring(node,node->v_strlen,1);
			ep = es_push();
			ep->type = ES_STATIC;
			ep->u.val1 = labelcnt++;
			ep->val2 = 0;
			break;
		 case TN_VALUE:
			ep = es_push();
			switch(node->format) {
				case NF_LEAF_CONST:
					ep->type = ES_CONST;
					ep->u.val1 = node->v_sval;
					break;
				case NF_LEAF_LABEL:
					ep->type = ES_LABEL;
					ep->u.str1 = node->v_name;
					ep->val2 = 0;
					break;
				case NF_LEAF_LOCAL:
					ep->type = ES_DREGOFFSET;
					ep->u.val1 = REG_FP;
					ep->val2 = node->v_offset;
					break;
				case NF_LEAF_PARAM:
					ep->type = ES_DREGOFFSET;
					ep->u.val1 = REG_AP;
					ep->val2 = node->v_offset;
					break;
				case NF_LEAF_STATIC:
					ep->type = ES_STATIC;
					ep->u.val1 = node->v_label;
					ep->val2 = 0;
					break;
				default:
					panic("unknown variable type");
			}
			break;
		case TN_LOGICOR:
		case TN_LOGICAND:
		case TN_LOGICNOT:
		case TN_EQUAL:
		case TN_NOTEQUAL:
		case TN_LESSTHAN:
		case TN_GREATERTHAN:
		case TN_LESSEQUAL:
		case TN_GREATEREQUAL:
			gen_logicexpr(node,0,0,0);
			break;
		case TN_FUNCTIONCALL:
			flushregs(1);
			n1 = node->link2;
			argcnt = 0;
			while(n1) {
				en.type = ES_PREDECR;
				en.u.val1 = REG_SP;
				if(!genexpr(n1->link2,&en,4)) {
					geninst("pushl",1,&expr_stack[es_pointer]);
					es_pop(1);
				}
				++argcnt;
				n1 = n1->link1;
			}
			genexpr(node->link1,0,4);
			en.type = ES_CONST;
			en.u.val1 = argcnt;
			geninst("calls",2,&en,&expr_stack[es_pointer]);
			es_pop(1);
			ep = es_push();
			ep->type = ES_REG;
			ep->u.val1 = 0;
			regmask &= ~1;
			break;
		case TN_ADD:
			return dobinop("add",node->link1,node->link2,rp,maxsize);
		case TN_SUBTRACT:
			genexpr(node->link1,0,maxsize);
			genexpr(node->link2,0,maxsize);
			stackfix(&expr_stack[es_pointer-1],maxsize);
			return dolrop("sub",rp,maxsize);
		case TN_MULTIPLY:
			return dobinop("mul",node->link1,node->link2,rp,maxsize);
		case TN_DIVIDE:
		case TN_MOD:
			siz = unxsize(node->link1);
			if(unxsize(node->link2) > siz) siz = unxsize(node->link2);
			if(!issigned(node))
				siz1 = 4;
			else if(siz > maxsize)
				siz1 = siz;
			else
				siz1 = maxsize;
			/*
			 *  Status quo:
			 *   siz     = max(size of left arg, size of right arg)
			 *   siz1    = size of operation to perform:
			 *				- always long if unsigned
			 *				- otherwise max(siz,maxsize)
			 *   maxsize = size wanted by caller
			 *
			 *  Special consideration needed when doing an unsigned divide
			 *  where either argument was already a longword.  This means that 
			 *  we can't rely on zero-extension to clear the sign bit, and thus
			 *  must generate a call to the library function udiv.
			 */
			if(!issigned(node) && siz == 4) {
				if(node->ntype == TN_MOD)
					panic("unsigned long modulo");
				flushregs(1);
				en.type = ES_PREDECR;
				en.u.val1 = REG_SP;
				if(!genexpr(node->link2,&en,siz)) {
					geninst("pushl",1,&expr_stack[es_pointer]);
					es_pop(1);
				}
				if(!genexpr(node->link1,&en,siz)) {
					geninst("pushl",1,&expr_stack[es_pointer]);
					es_pop(1);
				}
				printf("\tcalls\t$2,udiv\n");
				labelwaiting = 0;
				ep = es_push();
				ep->type = ES_REG;
				ep->u.val1 = 0;
				regmask &= ~1;
			} else if(node->ntype == TN_MOD) {
				genexpr(node->link1,0,siz1);
				genexpr(node->link2,0,siz1);
				registerize(siz1);
				en1 = expr_stack[es_pointer--];
				stackfix(&expr_stack[es_pointer],siz1);
				registerize(siz1);
				en = expr_stack[es_pointer--];
				ep = es_push();
				ep->type = ES_REG;
				ep->u.val1 = getreg();
				gensizop("div",siz1,3,3,&en1,&en,ep);
				gensizop("mul",siz1,2,2,ep,&en1);
				es_pop(1);
				if(rp) {
					gensizop("sub",maxsize,3,3,&en1,&en,rp);
					*es_push() = en1;
					*es_push() = en;
					es_pop(2);
					return 1;
				} else {
					gensizop("sub",siz1,2,2,&en1,&en);
					*es_push() = en;
					*es_push() = en1;
					es_pop(1);
				}
			} else {
				genexpr(node->link1,0,siz1);
				genexpr(node->link2,0,siz1);
				stackfix(&expr_stack[es_pointer-1],siz1);
				return dolrop("div",(siz1==maxsize)?rp:0,siz1);
			}
			break;
		case TN_DIVASSIGN:
			lsiz = siz = size(node->link1->vtype);
			if(unxsize(node->link2) > siz) siz = unxsize(node->link2);
			if(!issigned(node))
				siz1 = 4;
			else
				siz1 = siz;
			/*
			 *  Status quo:
			 *   siz     = max(size of left arg, size of right arg)
			 *   siz1    = size of operation to perform:
			 *				- always long if unsigned
			 *              - otherwise siz
			 */
			if(!issigned(node) && siz==4) {
				flushregs(1);
				regmask &= ~1;
				en.type = ES_PREDECR;
				en.u.val1 = REG_SP;
				if(!genexpr(node->link2,&en,siz)) {
					geninst("pushl",1,&expr_stack[es_pointer]);
					es_pop(1);
				}
				genexpr(node->link1,&en,lsiz);
				if(lsiz == 4)
					geninst("pushl",1,&expr_stack[es_pointer]);
				else {
					sprintf(ibuf,issigned(node->link1)?"cvt%cl":"movz%cl",
							"_bw"[lsiz]);
					geninst(ibuf,2,&expr_stack[es_pointer],&en);
				}
				printf("\tcalls\t$2,udiv\n");
				labelwaiting = 0;
				ep = es_push();
				ep->type = ES_REG;
				ep->u.val1 = 0;
				gensizop("mov",lsiz,0,2,ep,ep-1);
				es_swap();
				es_pop(1);
			} else if(siz1 == lsiz)
				genopassign(node,"div");
			else {
				genexpr(node->link1,0,siz1);
				ep = &expr_stack[es_pointer];
				genexpr(node->link2,0,siz1);
				stackfix(ep,siz1);
				en.type = ES_REG;
				en.u.val1 = getreg();
				sprintf(ibuf,issigned(node->link1)?"cvt%cl":"movz%cl",
						"_bw"[lsiz]);
				geninst(ibuf,2,ep,&en);
				gensizop("div",siz1,3,3,ep+1,&en,&en);
				gensizop("mov",lsiz,0,2,&en,ep);
				es_pop(2);
				*es_push() = en;
			}
			break;
		case TN_MODASSIGN:
			lsiz = siz = size(node->link1->vtype);
			if(unxsize(node->link2) > siz) siz = unxsize(node->link2);
			if(!issigned(node))
				siz1 = 4;
			else
				siz1 = siz;

			if(siz == 4 && !issigned(node->link1))
				panic("unsigned long %=");
			else {
				genexpr(node->link1,0,siz1);
				ep = &expr_stack[es_pointer];
				genexpr(node->link2,0,siz1);
				registerize(siz1);
				en1 = expr_stack[es_pointer--];

				stackfix(ep,siz1);
				en.type = ES_REG;
				en.u.val1 = getreg();
				if(lsiz < siz1) {
					sprintf(ibuf,issigned(node->link1)?"cvt%cl":"movz%cl",
							"_bw"[lsiz]);
					geninst(ibuf,2,ep,&en);
				} else
					gensizop("mov",lsiz,0,2,ep,&en);
				ep = es_push();
				ep->type = ES_REG;
				ep->u.val1 = getreg();
				gensizop("div",siz1,3,3,&en1,&en,ep);
				gensizop("mul",siz1,2,2,ep,&en1);
				es_pop(1);
				gensizop("sub",lsiz,3,3,&en1,&en,ep-1);
				*es_push() = en1;
				*es_push() = en;
				es_pop(2);
				return 0;
			}
			break;
		case TN_BINARYNOT:
			return comneg(COMPLEMENT,node->link1,rp,maxsize);
		case TN_BINARYOR:
			return dobinop("bis",node->link1,node->link2,rp,maxsize);
		case TN_BINARYAND:
			genexpr(node->link1,0,maxsize);
			comneg(COMPLEMENT,node->link2,0,maxsize);
			stackfix(&expr_stack[es_pointer-1],siz1);
			return dolrop("bic",rp,maxsize);
		case TN_EXOR:
			return dobinop("xor",node->link1,node->link2,rp,maxsize);
		case TN_UMINUS:
			return comneg(NEGATE,node->link1,rp,maxsize);
		case TN_LSHIFT:
			genexpr(node->link2,0,1);
			ep = &expr_stack[es_pointer];
			genexpr(node->link1,0,maxsize);
			stackfix(ep,siz1);
			if(rp && maxsize == 4) {
				if(unxsize(node->link1) != 4) registerize(maxsize);
				geninst("ashl",3,ep,ep+1,rp);
				es_pop(2);
				return 1;
			} else if(unxsize(node->link1) != 4) {
				registerize(maxsize);
				geninst("ashl",3,ep,ep+1,ep+1);
				es_swap();
				es_pop(1);
			} else {
				en.type = ES_REG;
				en.u.val1 = getreg();
				geninst("ashl",3,ep,ep+1,&en);
				es_pop(2);
				*es_push() = en;
			}
			break;
		case TN_LSHIFTASSIGN:
			siz = size(node->link1->vtype);
			genexpr(node->link2,0,1);
			ep = &expr_stack[es_pointer];
			genexpr(node->link1,0,siz);
			stackfix(ep,siz1);
			if(siz==4) {
				geninst("ashl",3,ep,ep+1,ep+1);
				es_swap();
				es_pop(1);
			} else {
				en.type = ES_REG;
				en.u.val1 = getreg();
				gensizop("mov",siz,0,2,ep+1,&en);
				geninst("ashl",3,ep,&en,&en);
				gensizop("mov",siz,0,2,&en,ep+1);
				es_pop(2);
				*es_push() = en;
			}
			break;
		case TN_RSHIFT:
			comneg(NEGATE,node->link2,0,1);
			ep = &expr_stack[es_pointer];
			genexpr(node->link1,0,4);
			stackfix(ep,siz1);
			if(unxsize(node->link1) == 4 && !issigned(node->link1))
				panic("unsigned long right shift");
			else if(rp && maxsize == 4) {
				geninst("ashl",3,ep,ep+1,rp);
				es_pop(2);
				return 1;
			} else if(isreg(ep+1)) {
				en.type = ES_REG;
				en.u.val1 = (ep+1)->u.val1;
				geninst("ashl",3,ep,ep+1,&en);
				es_pointer--;
				es_pop(1);
				*es_push() = en;
			} else if(isreg(ep)) {
				en.type = ES_REG;
				en.u.val1 = ep->u.val1;
				geninst("ashl",3,ep,ep+1,&en);
				es_pop(1);
				*ep = en;
			} else {
				en.type = ES_REG;
				en.u.val1 = getreg();
				geninst("ashl",3,ep,ep+1,&en);
				es_pop(2);
				*es_push() = en;
			}
			break;
		case TN_RSHIFTASSIGN:
			siz = size(node->link1->vtype);
			flag = issigned(node->link1);
			comneg(NEGATE,node->link2,0,1);
			ep = &expr_stack[es_pointer];
			genexpr(node->link1,0,siz);
			stackfix(ep,siz1);
			if(siz==4) {
				if(flag) {
					geninst("ashl",3,ep,ep+1,ep+1);
					es_swap();
					es_pop(1);
				} else
					panic("unsigned long right shift");
			} else {
				en.type = ES_REG;
				en.u.val1 = getreg();
				sprintf(ibuf,flag?"cvt%cl":"movz%cl","_bw"[siz]);
				geninst(ibuf,2,ep+1,&en);
				geninst("ashl",3,ep,&en,&en);
				gensizop("mov",siz,0,2,&en,ep+1);
				es_pop(2);
				*es_push() = en;
			}
			break;
		case TN_PREINCREMENT:
		case TN_PREDECREMENT:
			genexpr(node->link1,0,maxsize);
			doincdec(node);
			break;
		case TN_POSTINCREMENT:
		case TN_POSTDECREMENT:
			genexpr(node->link1,0,maxsize);
			ep = &expr_stack[es_pointer];
			if(rp) {
				gensizop("mov",size(node->link1->vtype),0,2,ep,rp);
				doincdec(node);
				es_pop(1);
				return 1;
			} else {
				es_push();
				(ep+1)->type = ES_REG;
				(ep+1)->u.val1 = getreg();
				gensizop("mov",size(node->link1->vtype),0,2,ep,ep+1);
				es_swap();
				doincdec(node);
				es_pop(1);
			}
			break;
		case TN_ADDASSIGN:
		case TN_SUBASSIGN:
			genexpr(node->link1,0,4);
			genexpr(node->link2,0,size(node->link1->vtype));
			ep = &expr_stack[es_pointer-1];
			t1 = skiprefnodes(node->link1->vtype);
			if(t1->type == TT_POINTERTO) {
				siz = size(t1->tn.next);
				multbyconst(ep+1,siz);
			}
			gensizop((node->ntype==TN_ADDASSIGN)?"add":"sub",
					 size(node->link1->vtype),2,2,ep+1,ep);
			es_pop(1);
			break;
		case TN_MULTASSIGN:
			genopassign(node,"mul");
			break;
		case TN_ORASSIGN:
			genopassign(node,"bis");
			break;
		case TN_EXORASSIGN:
			genopassign(node,"xor");
			break;
		case TN_ANDASSIGN:
			genexpr(node->link1,0,4);
			comneg(COMPLEMENT,node->link2,0,size(node->link1->vtype));
			ep = &expr_stack[es_pointer-1];
			stackfix(ep,siz1);
			gensizop("bic",size(node->link1->vtype),2,2,ep+1,ep);
			es_pop(1);
			break;
		case TN_CONDEXP:
			l1 = labelcnt++;
			l2 = labelcnt++;
			l3 = labelcnt++;
			gen_logicexpr(node->link1,l1,l2,l1);
			genlabel(l1);
			if(rp)
				flag = 1;
			else {
				flag = 0;
				rp = &en;
				en.type = ES_REG;
				en.u.val1 = getreg();
				regmask |= (1<<en.u.val1);
			}
			if(!genexpr(node->link2,rp,maxsize)) {
				if(!samething(&expr_stack[es_pointer],rp)) {
					*es_push() = *rp;
					gen_move(maxsize,2);
					--es_pointer;
				}
				es_pop(1);
			}
			en1.type = ES_STATIC;
			en1.u.val1 = l3;
			en1.val2 = 0;
			geninst("jbr",1,&en1);
			genlabel(l2);
			if(!genexpr(node->link3,rp,maxsize)) {
				if(!samething(&expr_stack[es_pointer],rp)) {
					*es_push() = *rp;
					gen_move(maxsize,2);
					--es_pointer;
				}
				es_pop(1);
			}
			genlabel(l3);
			if(!flag) {
				*es_push() = en;
				if(!(regmask & (1<<en.u.val1))) panic("cond expr reg bug");
				regmask &= ~(1<<en.u.val1);
			}
			return flag;
		case TN_VALUEAT:
			genexpr(node->link1,0,4);
			indirect();
			return 0;
			break;
		case TN_ADDRESSOF:
			genexpr(node->link1,0,4);
			return addressof(rp,maxsize);
			break;
		case TN_DOT:
			genexpr(node->v_dot_link,0,4);
			addoffset(node->v_dot_offset);
			return 0;
		case TN_PTRADD:
		case TN_PTRSUB:
			genexpr(node->link1,0,4);
			genexpr(node->link2,0,4);
			t1 = skiprefnodes(node->link1->vtype);
			if(t1->type != TT_POINTERTO) panic("pointer add on non-pointer");
			siz = size(t1->tn.next);
			if(skiprefnodes(node->link2->vtype)->type == TT_POINTERTO
			   && node->ntype == TN_PTRSUB) {
				if(siz==1)
					return dolrop("sub",rp,maxsize);
				else {
					dolrop("sub",0,4);
					return divbyconst(siz,(maxsize==4)?rp:0);
				}
			}
			return pointer_add((node->ntype==TN_PTRADD),siz,(maxsize==4)?rp:0);
			break;
		case TN_SUBSCRIPT:
			genexpr(node->link1,0,4);
			genexpr(node->link2,0,4);
			siz = size(node->vtype);
			do_subscript(siz);
			break;
		case TN_ASSIGN:
			genexpr(node->link2,0,size(node->link1->vtype));
			genexpr(node->link1,0,4);
			ep = &expr_stack[es_pointer-1];
			gen_move(size(node->link1->vtype),1);
			es_pop(1);
			break;
		case TN_EXPRESSION:
			while(node->link3) {
				eval_and_toss(node->link1);
				node = node->link3;
			}
			return genexpr(node->link1,rp,maxsize);
		default:
			panic("unknown tree node on genexpr");
	}
	return 0;
}

/*
 *  This function handles assignment/operator combinations, such as "+=".
 *  It takes as arguments a tree node representing the assignment operator,
 *  and a "root" opcode without size or number of argument extensions.
 *
 *  The assignment operator returns a value, which is left on top of the
 *  expression stack.  This value is the left operand of the assignment
 *  operator, after the assignment is done.
 */
genopassign(node,root)
struct treenode *node;
char *root;
{
	struct es_entry *ep;
	genexpr(node->link1,0,4);
	genexpr(node->link2,0,size(node->link1->vtype));
	ep = &expr_stack[es_pointer-1];
	gensizop(root,size(node->link1->vtype),2,2,ep+1,ep);
	es_pop(1);
}

/*
 *  This function handles increment/decrement operations on objects.  If the
 *  object incremented/decremented is a pointer, the size of the object is
 *  taken into account.
 */
doincdec(node)
struct treenode *node;
{
	int incflag = (node->ntype == TN_POSTINCREMENT ||
				   node->ntype == TN_PREINCREMENT);
	struct type * t1 = skiprefnodes(node->link1->vtype);
	struct es_entry en,*ep = &expr_stack[es_pointer];

	if(t1->type == TT_POINTERTO) {
		en.type = ES_CONST;
		en.u.val1 = size(t1->tn.next);
		geninst(incflag ? "addl2" : "subl2"	,2,&en,ep);
	} else
		gensizop(incflag ? "inc" : "dec", size(t1), 0, 1, ep);		
}

/*
 *  This routine generates a move instruction, not just for 32-bit and
 *  smaller ones but for bigger ones as well (using a movc3 instruction).
 *  But due to incomplete DEC documentation, it seems that a movc3 instruction,
 *  if interrupted, may clobber any or all of R0-R5.  So the routine flushes
 *  registers before doing this instruction.  The argument indicates how
 *  many entries at the top of the expression stack will not be needed any
 *  longer, and thus don't have to be flushed as it's OK to clobber any
 *  registers associated with them.
 */
gen_move(size,scrap)
int size;
{
	struct es_entry en,*ep = &expr_stack[es_pointer-1];
	switch(size) {
		case 1:
		case 2:
		case 4:
			gensizop("mov",size,0,2,ep,ep+1);
			break;
		default:
			en.type = ES_CONST;
			en.u.val1 = size;
			es_pointer -= scrap;
			flushregs(1);
			es_pointer += scrap;
			geninst("movc3",3,&en,ep,ep+1);
	}
}

/*
 *  This routine compares two addressing modes and sees if they point to/are
 *  the same object.  This is used to turn 3-operand instructions into
 *  2-operand instructions.  Note that (SP)+ and (SP)+ are NOT equal.
 */
int samething(thing1,thing2)
struct es_entry *thing1, *thing2;
{
	int t1 = thing1->type, t2 = thing2->type;
	switch(t1) {
		case ES_POSTINCR:
		case ES_POSTINCR+ES_DEFERRED:
		case ES_PREDECR:
			return 0;
		case ES_REG:
		case ES_CONST:
			return (t1 == t2 && thing1->u.val1 == thing2->u.val1);
		case ES_REG+ES_DEFERRED:
			if((t2 == ES_DREGOFFSET && thing1->val2 == 0)
			   || (t1 == t2 && thing1->val2 == thing2->val2))
				return(thing1->u.val1 == thing2->u.val1);
			else return 0;
		case ES_LABELADDR:
		case ES_LABEL:
		case ES_LABEL+ES_DEFERRED:
			return (t1 == t2 && thing1->val2 == thing2->val2
					&& !strcmp(thing1->u.str1,thing2->u.str1));
		case ES_DREGOFFSET:
			if(thing1->val2 == 0 && t2 == ES_REG+ES_DEFERRED)
				return(thing1->u.val1 == thing2->u.val1);
		case ES_DREGOFFSET+ES_DEFERRED:
		case ES_STATICADDR:
		case ES_STATIC:
		case ES_STATIC+ES_DEFERRED:
			return (t1 == t2 && thing1->u.val1 == thing2->u.val1
					&& thing1->val2 == thing2->val2);
	}
}

/*
 *  This function can be called instead of genexpr when it is known that
 *  the result of the expression won't be needed, as is the case in source
 *  lines such as this:
 *
 *   a=b;
 *
 *  The function takes advantage of this fact to generate better code for
 *  assignment operations (result is stored directly in the destination
 *  instead of cached in a register) and post increment/decrements (result
 *  is not loaded into a register before the inc/dec is done).
 */
eval_and_toss(node)
struct treenode *node;
{
	struct es_entry *ep;
	if(node->ntype == TN_EXPRESSION) {
		while(node) {
			eval_and_toss(node->link1);
			node = node->link3;
		}
	} else if(node->ntype == TN_ASSIGN) {
		genexpr(node->link1,0,4);
		ep = &expr_stack[es_pointer];
		if(!genexpr(node->link2,ep,size(node->link1->vtype))) {
			es_swap();
			gen_move(size(node->link1->vtype),2);
			es_pop(1);
		}
		es_pop(1);
	} else if(node->ntype == TN_POSTINCREMENT ||
			  node->ntype == TN_POSTDECREMENT) {
		genexpr(node->link1,0,4);
		doincdec(node);
		es_pop(1);
	} else {
		genexpr(node,0,1);
		es_pop(1);
	}
}

/*
 *  This computes the log base 2 of its argument.
 */
int log2(num)
long num;
{
	register long i=1;
	register int j;
	for(j=0;j<32;j++)
		if(i==num) return j;
		else i <<= 1;
	return -1;
}

/*
 *  This routine generates code to divide the top entry on the expression
 *  stack by a constant.  It tries to be reasonably efficient about this, 
 *  dividing a constant right away, and generating shift operations for
 *  division by a power of 2.  "siz" is the number to divide by.  "rp" is
 *  the usual, but "maxsize" is not given, instead the caller sets rp to
 *  zero when the sizes don't match.
 */
int divbyconst(siz,rp)
long siz;
struct es_entry *rp;
{
	struct es_entry en,*ep = &expr_stack[es_pointer];
	int nb = log2(siz);
	en.type = ES_CONST;
	if(siz==1)
		return 0;
	else if(ep->type == ES_CONST)
		ep->u.val1 /= siz;
	else if(rp) {
		if(nb>0) {
			en.u.val1 = -nb;
			geninst("ashl",3,&en,ep,rp);
		} else {
			en.u.val1 = siz;
			geninst("divl3",3,&en,ep,rp);
		}
		es_pop(1);
		return 1;
	} else {
		registerize(4);
		if(nb>0) {
			en.u.val1 = -nb;
			geninst("ashl",3,&en,ep,ep);
		} else {
			en.u.val1 = siz;
			geninst("divl2",2,&en,ep);
		}
	}
	return 0;
}

/*
 *  This routine multiplies an exprssion stack entry by a constant.  "ei" is
 *  the pointer to the stack entry, and "siz" is the constant to multiply
 *  by.  It tries to be reasonably efficient about this.
 */
multbyconst(ei,siz)
struct es_entry *ei;
long siz;
{
	struct es_entry en,en1;
	int nb = log2(siz);
	if(siz==1)
		return;
	else if(ei->type == ES_CONST) {
		ei->u.val1 *= siz;
	} else if(ei->type == ES_REG) {
		if(siz==2)
			geninst("addl2",2,ei,ei);
		else if(nb>0) {
			en.type = ES_CONST;
			en.u.val1 = nb;
			geninst("ashl",3,&en,ei,ei);
		} else {
			en.type = ES_CONST;
			en.u.val1 = siz;
			geninst("mull2",2,&en,ei);
		}
	} else {
		en.type = ES_CONST;
		en1.type = ES_REG;
		if(isreg(ei)) {
			en1.u.val1 = ei->u.val1;
			if(nb>0) {
				en.u.val1 = nb;
				geninst("ashl",3,&en,ei,&en1);
			} else {
				en.u.val1 = siz;
				geninst("mull3",3,&en,ei,&en1);
			}
			ei->type = ES_REG;
		} else {
			en1.u.val1 = getreg();
			if(nb>0) {
				en.u.val1 = nb;
				geninst("ashl",3,&en,ei,&en1);
			} else {
				en.u.val1 = siz;
				geninst("mull3",3,&en,ei,&en1);
			}
			*ei = en1;
		}
	}
}

/*
 *  This function implements pointer addition.  The first flag indicates
 *  whether an add or a subtract is to be done.  "siz" is the size of the
 *  type the pointer points to.  "rp" is the usual.
 *
 *  This is not much different from the code generation for any other
 *  noncommutative arithmetic operation.
 *
 *  The expression stack already contains the pointer and offset entries.
 */
pointer_add(addflag,siz,rp)
long siz;
int rp;
{
	struct es_entry en, *pe = &expr_stack[es_pointer-1];
	char *op = addflag ? "add" : "sub";
	multbyconst(pe+1,siz);
	if(rp) {
		if(samething(pe,rp)) gensizop(op,4,2,2,pe+1,pe);
		else gensizop(op,4,3,3,pe+1,pe,rp);
		es_pop(2);
		return 1;
	} else if(pe->type == ES_REG) {
		gensizop(op,4,2,2,pe+1,pe);
		es_pop(1);
	} else if((pe+1)->type == ES_REG && addflag) {
		gensizop(op,4,2,2,pe,pe+1);
		es_swap();
		es_pop(1);
	} else {
		en.type = ES_REG;
		if(isreg(pe)) {
			en.u.val1 = pe->u.val1;
			gensizop(op,4,3,3,pe+1,pe,&en);
			es_pop(1);
			*pe = en;
		} else if(isreg(pe+1)) {
			en.u.val1 = (pe+1)->u.val1;
			gensizop(op,4,3,3,pe+1,pe,&en);
			es_swap();
			es_pop(1);
			*pe = en;
		} else {
			en.u.val1 = getreg();
			gensizop(op,4,3,3,pe+1,pe,&en);
			es_pop(2);
			*es_push() = en;
		}
	}
	return 0;
}

/*
 *  This implements an array subscript (i.e. a pointer add but to an effective
 *  address).  More optimization could be done here but I ran out of steam
 *  and hacked it as a combination of already written routines.
 *
 *  expr_stack:        top: subscript (integer)
 *              1 from top: array base (effective address)
 *
 *  siz: size of one array element.
 */
do_subscript(siz)
long siz;
{
	es_swap();
	addressof();
	es_swap();
	pointer_add(1,siz,0);
	indirect();
}
