#include "exprstack.h"
#include "nodetypes.h"
#include "nodes.h"
#include "symtab.h"
#include "error.h"
#include "types.h"

static char *sccsid = "@(#) logicexpr.c 1.22@(#)";

extern int labelcnt;
extern struct es_entry expr_stack[];
extern int es_pointer;
extern struct es_entry *es_push();
extern void es_pop();

/*
 *  This computes the "unextended size" of an expression.  It looks in, or past
 *  any extend nodes at the top of the tree to find the true size of the 
 *  object returned (i.e. how many significant bits are in it).
 */
int unxsize(node)
struct treenode *node;
{
	switch(node->ntype) {
		case TN_UBYTETOLONG:
		case TN_SBYTETOLONG:
			return 1;
		case TN_UWORDTOLONG:
		case TN_SWORDTOLONG:
			return 2;
		default:
			return size(node->vtype);
	}
}

/*
 *  This determines whether a type is a signed or an unsigned one.
 */
int issigned(node)
struct treenode *node;
{
	extern struct type *skiprefnodes();
	struct type *t = skiprefnodes(node->vtype);
	if(t->type == TT_POINTERTO) return 0;
	else if(t->type == TT_ENUMHEADER) return 1;
	else if(t->type == TT_BASICTYPE)
		return !(t->tu.flags & TF_UNSIGNED);
	else
		panic("issigned: bad arithmetic type");
}

/*
 *  This is the third (after codegen and genexpr) most important function in
 *  the code generator.  Like the others, it is recursive.  It generates 
 *  code for a short-circuited boolean expression at "node", generating some
 *  sort of branch to "ltrue" if it succeeds, "lfalse" if it fails.  If the
 *  caller intends to generate a label right after this returns, it may
 *  indicate this by setting "lnext" to this label so this function can
 *  eliminate branches to locations directly following.
 *
 *  Note:  If this function is called by codegen or itself, then ltrue and
 *         lfalse are set, and the caller expects a jump to those locations.
 *         If it is called by genexpr, then ltrue and lfalse are zero, and
 *         it must return 0 or 1, since the value of the boolean expression
 *         will actually be used.
 */
gen_logicexpr(node,ltrue,lfalse,lnext)
register struct treenode *node;
int ltrue,lfalse,lnext;
{
	int flag = 0;
	struct es_entry en, *ep;
	int lab,lt,lf;
	int sgn1,sgn2,siz1,siz2;

	if(!ltrue) {
		flag = 1;
		ltrue = labelcnt++;
		lfalse = labelcnt++;
		lnext = ltrue;
	}
	switch(node->ntype) {
		case TN_LOGICNOT:
			gen_logicexpr(node->link1,lfalse,ltrue,lnext);
			break;
		case TN_LOGICOR:
			lab = labelcnt++;
			gen_logicexpr(node->link1,ltrue,lab,lab);
			genlabel(lab);
			gen_logicexpr(node->link2,ltrue,lfalse,lnext);
			break;
		case TN_LOGICAND:
			lab = labelcnt++;
			gen_logicexpr(node->link1,lab,lfalse,lab);
			genlabel(lab);
			gen_logicexpr(node->link2,ltrue,lfalse,lnext);
			break;
		case TN_EXPRESSION:
			while(node->link3) {
				eval_and_toss(node->link1);
				node = node->link3;
			}
			gen_logicexpr(node->link1,ltrue,lfalse,lnext);
			break;
		case TN_EQUAL:
		case TN_NOTEQUAL:
		case TN_LESSTHAN:
		case TN_GREATERTHAN:
		case TN_LESSEQUAL:
		case TN_GREATEREQUAL:
			siz1 = unxsize(node->link1);
			siz2 = unxsize(node->link2);

			sgn1 = issigned(node->link1);
			sgn2 = issigned(node->link2);

			if(siz1<siz2) {
				siz1 = siz2;
				sgn1 = sgn2;
			} else if(siz1 == siz2)
				sgn1 = (sgn1 && sgn2);

			genexpr(node->link1,0,siz1);
			genexpr(node->link2,0,siz1);

			ep = &expr_stack[es_pointer-1];
			stackfix(ep,siz1);
			gensizop("cmp",siz1,0,2,ep,ep+1);
			es_pop(2);
			switch(node->ntype) { 
				case TN_EQUAL:
					gencjmp("jeql","jneq",ltrue,lfalse,lnext);
					break;
			    case TN_NOTEQUAL:
					gencjmp("jeql","jneq",lfalse,ltrue,lnext);
					break;
				case TN_LESSTHAN:
					if(sgn1)
						gencjmp("jlss","jgeq",ltrue,lfalse,lnext);
					else
						gencjmp("jlssu","jgequ",ltrue,lfalse,lnext);
					break;
				case TN_GREATERTHAN:
					if(sgn1)
						gencjmp("jgtr","jleq",ltrue,lfalse,lnext);
					else
						gencjmp("jgtru","jlequ",ltrue,lfalse,lnext);
					break;
				case TN_LESSEQUAL:
					if(sgn1)
						gencjmp("jgtr","jleq",lfalse,ltrue,lnext);
					else
						gencjmp("jgtru","jlequ",lfalse,ltrue,lnext);
					break;
				case TN_GREATEREQUAL:
					if(sgn1)
						gencjmp("jlss","jgeq",lfalse,ltrue,lnext);
					else
						gencjmp("jlssu","jgequ",lfalse,ltrue,lnext);
					break;
			}
			break;
		default:
			siz1 = unxsize(node);
			genexpr(node,0,siz1);
			gensizop("tst",siz1,0,1,&expr_stack[es_pointer]);
			es_pop(1);
			gencjmp("jneq","jeql",ltrue,lfalse,lnext);
			break;
	}
	if(flag) {
		ep = es_push();
		ep->type = ES_REG;
		ep->u.val1 = getreg();
		genlabel(ltrue);
		en.type = ES_CONST;
		en.u.val1 = 1;
		geninst("movl",2,&en,ep);
		en.type = ES_STATIC;
		en.u.val1 = labelcnt;
		en.val2 = 0;
		geninst("brb",1,&en);
		genlabel(lfalse);
		geninst("clrl",1,ep);
		genlabel(labelcnt++);
	}
}

/*
 *  This function generates a conditional jump.  In order to do this
 *  efficiently, it takes the following operands:
 *
 *   optrue:   opcode which jumps when condition is true
 *   opfalse:  opcode which jumps when condition is false
 *   ltrue:    label to go to when true
 *   lfalse:   label to go to when false
 *   lnext:    next consecutive label
 *
 *  This allows "branches around branches" to be eliminated completely on
 *  a per-statement basis.  They still occur in the case of compounded
 *  statements, such as "if(cond) break".
 */
gencjmp(optrue,opfalse,ltrue,lfalse,lnext)
char *optrue,*opfalse;
int ltrue,lfalse,lnext;
{
	struct es_entry en;
	if(ltrue == lnext) {
		en.type = ES_STATIC;
		en.u.val1 = lfalse;
		en.val2 = 0;
		geninst(opfalse,1,&en);
	} else if(lfalse == lnext) {
		en.type = ES_STATIC;
		en.u.val1 = ltrue;
		en.val2 = 0;
		geninst(optrue,1,&en);
	} else {
		en.type = ES_STATIC;
		en.u.val1 = ltrue;
		en.val2 = 0;
		geninst(optrue,1,&en);
		en.u.val1 = lfalse;
		geninst("jbr",1,&en);
	}
}
