#include "def.h"
#include "nodes.h"
#include "symtab.h"
#include "types.h"
#include "error.h"
#include "nodetypes.h"

static char *sccsid = "@(#) nodes1.c 1.14@(#)";

/*
 *  The basic tree node construction function.  This one manufactures most
 *  "plain" tree nodes, and allocates the underlying node structure of more
 *  complex types.
 *
 *   type:      The value for the "ntype" field
 *   l1,l2,l3:  Pointers to up to three child nodes
 */
struct treenode *make_plain(type,l1,l2,l3)
int type;
struct treenode *l1, *l2, *l3;
{
	struct treenode *node=(struct treenode *)safealloc(sizeof(struct treenode));
	node->ntype = type;
	node->link1 = l1;
	node->link2 = l2;
	node->link3 = l3;
	node->vtype = 0;
	node->lineno = linecnt;
	node->format = NF_INNER;
	return node;
}

/*
 *  make_bit32:  take a tree node and process it into something which can fit
 *               into a 32-bit machine word.
 *
 *  enum: 				leave alone
 *  8/16-bit integral:  extend to 32-bit integral
 *  pointer:			leave alone
 *  void:				scrap and replace with constant '0'
 *  array:				turn into pointer to first element
 *  function:			turn into pointer to function
 *  struct/union:		scrap and replace with constant '0'
 */
struct treenode *make_bit32(node)
struct treenode *node;
{
	extern struct treenode *make_const();
	extern struct type *maketypenode();
	struct type *t1 = skiprefnodes(node->vtype);
	switch(t1->type) {
		case TT_ENUMHEADER:
		case TT_POINTERTO:
			return node;
		case TT_ARRAYOF:
			node = make_plain(TN_ADDRESSOF,node,0,0);
			t1 = t1->tn.next;
			maketypenode(&t1,TT_POINTERTO);
			node->vtype = t1;
			node->vtype->refcount++;
			return node;
		case TT_FUNCTIONRETURNING:
			node = make_plain(TN_ADDRESSOF,node,0,0);
			maketypenode(&t1,TT_FUNCTIONRETURNING);
			node->vtype = t1;
			node->vtype->refcount++;
			return node;
		case TT_BASICTYPE:
			if(t1->tu.flags & TF_VOID) {
				err(VOIDEXPR);
				kill_tree(node);
				return make_const(0,0,0);
			}
			extend(&node);
			return node;
		case TT_STRUCTHEADER:
		case TT_UNIONHEADER:
			err(STRUCTNOTVALID);
			kill_tree(node);
			return make_const(0,0,0);
		default:
			panic("bad type in make_bit32");
	}
}

/*
 *  pointermatch:  take two type chains and return true if they are both
 *                 pointers to the same type, or if one is a pointer to void
 *				   and the other is any sort of pointer.  return false if
 *				   one or both is not a pointer, or if the above is not met.
 */
int pointermatch(t1,t2)
struct type *t1,*t2;
{
	t1 = skiprefnodes(t1);
	t2 = skiprefnodes(t2);
	if(t1->type != TT_POINTERTO || t2->type != TT_POINTERTO) return 0;
	if((t1->tn.next->type == TT_BASICTYPE && t1->tn.next->tu.flags & TF_VOID)
	  ||(t2->tn.next->type == TT_BASICTYPE && t2->tn.next->tu.flags & TF_VOID))
		return 1;
	return typematch(t1,t2);
}

/*
 *  integerpromote:  take a tree node and turn it into a signed or unsigned
 *					 32-bit integer.  this means processing it as follows:
 *
 *  enum: 				convert to integer
 *  8/16-bit integral:  extend to 32-bit
 *  pointer:			convert to integer
 *  void:				scrap and replace with constant '0'
 *  array:				turn into pointer to first element, then handle as
 *						pointer (see above)
 *  function:			turn into pointer to function, then handle as pointer
 *						(see above)
 *  struct/union:		scrap and replace with constant '0'
 */
struct treenode *integerpromote(node)
struct treenode *node;
{
	extern struct type *make_int();
	struct type *t1 = skiprefnodes(node->vtype);
	switch(t1->type) {
		case TT_ENUMHEADER:
			freetype(node->vtype);
			node->vtype = make_int();
			return node;
		case TT_BASICTYPE:
		case TT_STRUCTHEADER:
		case TT_UNIONHEADER:
			return make_bit32(node);
		case TT_ARRAYOF:
		case TT_FUNCTIONRETURNING:
			node = make_bit32(node);
		case TT_POINTERTO:
			freetype(node->vtype);
			node->vtype = make_int();
			return node;
		default:
			panic("bad type in integerpromote");
	}
}

/*
 *  make_int:  obtain a pointer to an integer type.  the reference count in
 *			   the node to which a pointer is returned is incremented
 *			   once.  where this integer node comes from, nooobody knows...
 */
struct type *make_int()
{
	static struct type *intnode = 0;
	if(!intnode) {
		intnode = (struct type *)safealloc(sizeof(struct type));
		intnode->type = TT_BASICTYPE;
		intnode->refcount = 0;
		intnode->tu.flags = TF_BIT32;
	}
	intnode->refcount++;
	return intnode;
}


/*
 *  Tree node construction function for various conditional things.  This
 *  verifies that the boolean expressions for FOR, WHILE, IF, ||, &&, !, etc
 *  are valid, ensures that the latter three return "int", and handles the
 *  type checking for a conditional (?:) expression.
 *
 *  When all is said and done, this function returns with a tree node of the
 *  appropriate type.
 */
struct treenode *make_cond(type,l1,l2,l3)
int type;
struct treenode *l1, *l2, *l3;
{
	struct type *t2,*t3,*tf;
	struct treenode *node;
	extern struct type *make_sign();
	switch(type) {
		case TN_IF:
		case TN_IF_ELSE:
		case TN_WHILE:
			return make_plain(type,l1,make_bit32(l2),l3);
		case TN_FOREXP:
			if(l2) l2 = make_bit32(l2);
			return make_plain(type,l1,l2,l3);
		case TN_DOWHILE:
			return make_plain(type,l1,l2,make_bit32(l3));
		case TN_CONDEXP:
			l1 = make_bit32(l1);
			l2 = make_bit32(l2);
			l3 = make_bit32(l3);
			t2 = skiprefnodes(l2->vtype);
			t3 = skiprefnodes(l3->vtype);
			if( t2->type == TT_ENUMHEADER && typematch(t2,t3)) {
				tf = t2;
			} else if(t2->type == TT_POINTERTO) {
				if( pointermatch(t2,t3)
					||(l3->format==NF_LEAF_CONST&&!l3->v_uval)) {
					tf = t2;
				} else if(t3->type == TT_POINTERTO) {
					err(PTRPTRCONV);
					tf = t2;
				} else if(l3->format == NF_LEAF_CONST && !l3->v_uval) {
					err(PTRINTCONV);
					l3 = integerpromote(l3);
					tf = l3->vtype;
				}
			} else if(t3->type==TT_POINTERTO) {
				if(l2->format==NF_LEAF_CONST & !l2->v_uval) {
					tf = t3;
				} else {
					err(PTRINTCONV);
					l2 = integerpromote(l2);
					l3 = integerpromote(l3);
					tf = make_sign(l2->vtype,l3->vtype);
				}
			} else {
				l2 = integerpromote(l2);
				l3 = integerpromote(l3);
				tf = make_sign(l2->vtype,l3->vtype);
			}
			node = make_plain(type,l1,l2,l3);
			node->vtype = tf;
			node->vtype->refcount++;
			return node;
		case TN_LOGICAND:
		case TN_LOGICOR:
			l2 = make_bit32(l2);
		case TN_LOGICNOT:
			l1 = make_bit32(l1);
			node = make_plain(type,l1,l2,l3);
			node->vtype = make_int();
			return node;
		default:
			panic("bad node type in make_cond");
	}
}

/*
 *  Tree node constructor for a function argument node.  Not much here now,
 *  we just ensure that 32-bit integer, pointer, or enumerated types are be
 *  passed.  This would be much more complex if function arguments were handled
 *  ANSI style.
 */
struct treenode *make_arg(type,l1,l2)
int type;
struct treenode *l1,*l2;
{
	return make_plain(type,l1,make_bit32(l2),0);
}


/*
 *  Tree node constructor and type checker for increment and decrement
 *  (++/--) operators.
 */
struct treenode *make_incdec(type,l1)
int type;
struct treenode *l1;
{
	struct type *t1 = skiprefnodes(l1->vtype);
	struct treenode *node;
	if( islval(l1) ) {
		if( t1->type == TT_POINTERTO ) {
			if(skiprefnodes(t1->tn.next)->type == TT_FUNCTIONRETURNING) {
				err(INVALIDTYPEFOROP,(type == TN_POSTINCREMENT ||
					type == TN_PREINCREMENT) ? "++" : "--" );
				node = l1;
			} else {
				node = make_plain(type,l1,0,0);
				node->vtype = node->link1->vtype;
				node->vtype->refcount++;
			}	
		} else if(t1->type == TT_BASICTYPE || t1->type == TT_ENUMHEADER) {
			node = make_plain(type,l1,0,0);
			node->vtype = node->link1->vtype;
			node->vtype->refcount++;
			node = integerpromote(node);
		} else {
			err(INVALIDTYPEFOROP,(type == TN_POSTINCREMENT ||
				type == TN_PREINCREMENT) ? "++" : "--" );
			node = l1;
		}
	} else {
		err(LVALUEREQD);
		node = l1;
	}
	return node;
}

/*
 *  Tree node constructor for identifies.  This takes an identifier
 *  node as an argument, and looks it up in the symbol table.  If found, it
 *  uses the symbol table information to generate a tree node of the
 *  appropriate type.  This may be a constant node if the identifier is an
 *  enumerator.
 *
 *  If the symbol table entry is not found, then one is generated with the
 *  type set to "global integer", an error is printed, and code generation
 *  continues.
 */
struct treenode *make_var(id)
struct treenode *id;
{
	extern struct treenode *make_const();
	extern struct symbol *find_next();

	struct treenode *node=(struct treenode *)safealloc(sizeof(struct treenode));
	struct symbol *sym = find_sym(id->v_name,ET_ARGUMENT|ET_LOCAL|ET_GLOBAL,0);

	if(!sym) {
		sym = find_sym(0,ET_ENUMDEF,0);
		while(sym) {
			struct enumerator *en;
			if(sym->type->type != TT_ENUMHEADER) panic("bad enum sym");
			en = sym->type->tn.enumlist;
			while(en) {
				if(!strcmp(id->v_name,en->name)) {
					node->format = NF_LEAF_CONST;
					node->v_uval = en->value;
					node->vtype = sym->type;
					node->vtype->refcount++;
					node->lineno = linecnt;
					node->ntype = TN_VALUE;
					return node;
				}
				en = en->next;
			}
			sym = find_next(sym->next,0,ET_ENUMDEF,0);
		}
		errl(UNDEFINED,id,id->v_name);
		sym = make_sym(blocklevel,id->v_name,ET_GLOBAL);
		sym->storclass = TN_EXTERN;
		sym->type = (struct type *) safealloc(sizeof(struct type));
		sym->type->type = TT_BASICTYPE;
		sym->type->refcount = 1;
		sym->type->tu.flags = TF_BIT32;
	}
	node->ntype = TN_VALUE;
	node->lineno = linecnt;
	node->vtype = sym->type;
	node->vtype->refcount++;
	switch(sym->entrytype) {
		case ET_GLOBAL:
			node->format = NF_LEAF_LABEL;
			node->v_name = save_str(sym->name);
			break;
		case ET_ARGUMENT:
			node->format = NF_LEAF_PARAM;
			node->v_offset = sym->offset;
			break;
		case ET_LOCAL:
			if(sym->storclass == TN_STATICLOCAL) 
				node->format = NF_LEAF_STATIC;
			else
				node->format = NF_LEAF_LOCAL;
			node->v_label = sym->offset;
			break;
	}
	kill_tree(id);
	return node;
}

/*
 *  Node construction function for a function call.  This takes two node
 *  arguments, one referring to the function and the other to the parameter
 *  list.  If the function is given as an identifer, and the identifier is
 *  not in the symbol table, then the function is magically declared as
 *  "external fuction returning int" at the global level.
 *
 *  In any case, the type of the function must be "function" or "pointer to
 *  function".  In the latter case, a "TN_VALUEAT" node is inserted ahead
 *  of the function automatically to remove the "pointer to" part.
 */
struct treenode *make_functioncall(node1,node2)
struct treenode *node1,*node2;
{
	struct type *tnode;
	extern struct type *maketypenode();
	struct symbol *sym;

	struct treenode *new = (struct treenode*)safealloc(sizeof(struct treenode));

	new->ntype = TN_FUNCTIONCALL;
	new->format = NF_INNER;
	new->lineno = linecnt;
	new->link2 = node2;
	new->link3 = NULL;

	if(node1->ntype == TN_ID) {
		if(!find_sym(node1->v_str, ET_ARGUMENT|ET_LOCAL|ET_GLOBAL, 0)) {
			sym = make_sym(0,node1->v_str,ET_GLOBAL);
			sym->storclass = TN_EXTERN;
			tnode = (struct type *) safealloc(sizeof(struct type));
			tnode->type = TT_BASICTYPE;
			tnode->refcount = 1;
			tnode->tu.flags = TF_BIT32;
			maketypenode(&tnode,TT_FUNCTIONRETURNING);
			tnode->tu.flags = 0;
			sym->type = tnode;
		}
		node1 = make_var(node1);
	}
	if(node1->vtype->type == TT_POINTERTO)
		node1 = make_inner(TN_VALUEAT,node1,0,0);
	if(node1->vtype->type != TT_FUNCTIONRETURNING) {
		err(CALLTONONFUNCTION);
		free(new);
		if(node2) kill_tree(node2);
		return node1;
	}
	new->vtype = node1->vtype->tn.next;
	new->vtype->refcount++;
	new->link1 = node1;
	return new;
}
