#include "def.h"
#include "nodes.h"
#include "symtab.h"
#include "types.h"
#include "error.h"
#include "nodetypes.h"

static char *sccsid = "@(#) nodes2.c 1.15@(#)";

/*
 *  This is the master tree node construction function, used for all the things
 *  which don't have their own special function but which require more work
 *  than make_plain does.
 *
 *  Most of the arithmetic stuff is just dispatched to the appropriate type
 *  checker functions.  This is a bit inefficient; it would have been better
 *  to call these from the grammar directly.
 */
struct treenode *make_inner(type,l1,l2,l3)
int type;
struct treenode *l1,*l2,*l3;
{
	struct treenode *node, *eval_assign(), *check_type();
	switch(type) {
		/* check arith is called directly
		*/
		case TN_ADD: case TN_SUBTRACT: case TN_MULTIPLY: case TN_DIVIDE:
		case TN_MOD: case TN_UMINUS:
		case TN_LSHIFT: case TN_RSHIFT:
		case TN_BINARYNOT: case TN_BINARYAND: case TN_BINARYOR: case TN_EXOR:
			return check_arith(type,l1,l2,l3);

		/* make_cond is called instead
		*/
		case TN_EQUAL: case TN_NOTEQUAL:
		case TN_GREATEREQUAL: case TN_GREATERTHAN: case TN_LESSEQUAL:
		case TN_LESSTHAN:
			return check_cond(type,l1,l2,l3);

		case TN_ADDRESSOF:
			switch( l1->vtype->type ) {
				case TT_FUNCTIONRETURNING: case TT_ARRAYOF:
					err(ADDRESSIGNORED);
					return l1;
				default:
					if(islval(l1)) {
						return make_addressof( l1 );
					} else {
						err(NOADDRESS);
						return l1;
					}
			}
		case TN_VALUEAT:
			{
				struct type *t1 = skiprefnodes(l1->vtype);
				switch( t1->type ) {
					case TT_ARRAYOF:
						l1 = make_addressof( l1 );
					case TT_POINTERTO:
						node = make_plain(type,l1,l2,l3);
						node->vtype = t1->tn.next;
						t1->tn.next->refcount++;
						node->lineno = node->link1->lineno;
						return node;
					case TT_FUNCTIONRETURNING:
						return l1;
					default:
						err(POINTERREQD);
						return l1;
				}
			}
		case TN_EXPRESSION:
			node = make_plain(type,l1,l2,l3);
			node->vtype = l1->vtype;
			node->vtype->refcount++;
			node->lineno = node->link1->lineno;
			return node;
		case TN_SWITCH:
			return make_plain(type,l1,make_bit32(l2),l3);
		case TN_RETURN:
			if(l2) return make_plain(type,l1,make_bit32(l2),l3);
			else return make_plain(type,l1,l2,l3);
		case TN_ARROW:
			l1 = make_inner(TN_VALUEAT,l1,0,0);
		case TN_DOT:
			{
				struct type *t1 = skiprefnodes(l1->vtype);
				struct member *mp;

				if(t1->type != TT_STRUCTHEADER && t1->type != TT_UNIONHEADER) {
					err(INVALIDTYPEFOROP,(type==TN_DOT)?".":"->");
					kill_tree(l2);
					return l1;
				}
				if(!t1->tn.members) {
					err(INCOMPLETETYPE);
					kill_tree(l2);
					return l1;
				}
				mp = t1->tn.members;
				while(mp) {
					if(!strcmp(l2->v_str,mp->name)) break;
					mp = mp->next;
				}
				if(!mp) {
					err(NOTMEMBER,l2->v_str);
					mp = t1->tn.members;
				}
				node = (struct treenode *) safealloc(sizeof(struct treenode));
				node->ntype = TN_DOT;
				node->format = NF_INNER_DOT;
				node->lineno = linecnt;
				node->v_dot_link = l1;

				node->vtype = mp->type;
				mp->type->refcount++;
				node->v_dot_offset = mp->offset;
				kill_tree(l2);
				return node;
			}
		case TN_SUBSCRIPT:
			{
				struct type *t1 = skiprefnodes( l1->vtype ),
							*t2 = skiprefnodes( l2->vtype );

				/* check for right types that allow subscripts */

				if( (t2->type == TT_POINTERTO || t2->type == TT_ARRAYOF)
					&& t1->type == TT_BASICTYPE) {
					struct treenode *lt = l1;
					struct type *tt = t1;
					l1 = l2; l2 = lt;
					t1 = t2; t2 = tt;
				} else if( (t1->type != TT_POINTERTO && t1->type != TT_ARRAYOF)
					|| t2->type != TT_BASICTYPE) {
					err(INVALIDSUBTYPE);
					kill_tree(l2);
					return l1;
				}

				if(t1->type == TT_POINTERTO) {
					l1 = make_plain(TN_VALUEAT,l1,0,0);
					l1->vtype = l1->link1->vtype;
					l1->vtype->refcount++;
					l1->lineno = l1->link1->lineno;
				}

				/*
				 * get rid of constant array node values (turn into dot nodes)
				 * (only if expression node is the only one)
				 */

				if( l2->format == NF_LEAF_CONST ) {
					node =(struct treenode *)safealloc(sizeof(struct treenode));
					node->ntype = TN_DOT;
					node->format = NF_INNER_DOT;
					node->lineno = linecnt;
					node->v_dot_link = l1;

					node->vtype = skiprefnodes( t1->tn.next );
					node->vtype->refcount++;

					if( l2->vtype->tu.flags & TF_UNSIGNED )
						node->v_dot_offset = l2->v_uval;
					else
						node->v_dot_offset = l2->v_uval;

					if(l1->vtype->type == TT_POINTERTO) 
						node->v_dot_offset *= size(l1->vtype->tn.next);
					else
						node->v_dot_offset *= l1->vtype->tu.array.sizeeach;
					kill_tree(l2);
					return node;
				}
				extend(&l2);
				node = make_plain(type,l1,l2,l3);
				node->vtype = (struct type *) safealloc(sizeof(struct type));
				node->vtype = skiprefnodes( t1->tn.next );
				node->vtype->refcount++;
				node->lineno = linecnt;
				return node;
			}
		case TN_CAST:
			{
				int sc;
				struct type *t1, *t2, *t3 = skiprefnodes(l2->vtype);
				struct treenode *p1;
				/*
				 *  Process the type expression, and build a type chain
				 *  rooted at t1.
				 */
				t2 = process_declaration_specs(l1->link1,&sc,0);
				p1 = l1->link2;
				t1 = buildtypechain(&p1,t2,0);
				freetype(t2);

				if(t3->type == TT_ARRAYOF || t3->type == TT_FUNCTIONRETURNING)
					l2 = make_addressof(l2);

				if(size(t1) > size(l2->vtype) && l2->format != NF_LEAF_CONST) {
					p1 = l2;
					extend(&p1);
					freetype(p1->vtype);
					p1->vtype = t1;
					return p1;
				} else {
					freetype(l2->vtype);
					l2->vtype = t1;
					return l2;
				}
			}
		default:
			panic("unsupported node in make_inner\n");
	}
	panic("make_inner");
	/* NOTREACHED */
}
