#include "exprstack.h"
#include "nodetypes.h"
#include "nodes.h"
#include "symtab.h"
#include "error.h"
#include "def.h"

static char *sccsid = "@(#) switch.c 1.22@(#)";

extern int labelcnt;
extern struct es_entry *es_push();
extern void es_pop();
extern struct es_entry expr_stack[];
extern int es_pointer;
extern int labelcnt;

struct treenode *head;
struct treenode **current;
int defaultlabel;
int casecount;

/*
 *  An array of these structures is dynamically allocated to store the case
 *  values and corresponding destination labels.
 */
struct caseentry {
	long value;
	int label;
};

/*
 *  This is the subroutine used by qsort() to sort the case labels in ascending
 *  order by value.
 */
int casecmp(case1,case2)
struct caseentry *case1, *case2;
{
	return case1->value - case2->value;
}

/*
 *  This function takes the top of the expression stack and forces it into
 *  a register.  "siz" is used to determine the size of "mov" instruction
 *  to generate.
 */
registerize(siz)
int siz;
{
	struct es_entry en, *ep = &expr_stack[es_pointer];
	stackfix(ep,0);
	if(ep->type != ES_REG) {
		if((ep->type == ES_REG+ES_DEFERRED || ep->type == ES_DREGOFFSET
		   || ep->type == ES_DREGOFFSET+ES_DEFERRED) && ep->u.val1 < 6)
			en.u.val1 = ep->u.val1;
		else
			en.u.val1 = getreg();
		en.type = ES_REG;
		gensizop("mov",siz,0,2,ep,&en);
		*ep = en;
	}
}

/*
 *  This function implements the "switch" operation.  A switch statement is
 *  subject to "break" but not to "continue" operations, so the place to jump
 *  to on a "continue" is passed down from above in the argument "contlabel".
 *
 */
handle_switch(node,contlabel)
struct treenode *node;
{
	register struct treenode *tp;
	register struct caseentry *casearray, *cptr, *cptr2;
	long casemin,casemax,tablesize,lastval;
	register int i;
	int baselabel,endlabel,lab;
	struct es_entry en1,en2,en3,*ep;

	current = &head;
	head = 0;
	defaultlabel = 0;
	casecount = 0;
	endlabel = labelcnt++;

	/*
	 *  Gather up all the case labels inside the statement following the
	 *  switch, by stringing them all together.
	 */
	gathercases(node->link3);
	/*
	 *  Set the default label to the exit point of the switch statement if
	 *  no "default" label was provided.
	 */
	if(!defaultlabel) defaultlabel = endlabel;

	if(casecount) {
		/*
		 *  Now that we know how many cases there are, we can dynamically
		 *  allocate an array to hold them.
		 */
		cptr = casearray = (struct caseentry *)
			safealloc(casecount * sizeof(struct caseentry));
		/*
		 *  Traverse the string of case labels built earlier, undoing it as
		 *  we go, (and returning the tree to a safe state).  Store the
		 *  gathered information in the case array, and also find the maximum
		 *  and minimum case values.
		 */
		casemin = 0x7fffffff;
		casemax = 0x80000000;
		while(head) {
			tp = head;
			head = tp->link3;
			tp->link3 = 0;
			cptr->label = tp->link2->v_sval;
			cptr->value = get_const(tp->link1);
			if(cptr->value > casemax) casemax = cptr->value;
			if(cptr->value < casemin) casemin = cptr->value;
			++cptr;
		}
		/*
		 *  If the switch is implemented as a VAX "case" instruction, then
		 *  this is how many entries the table will have.
		 */
		tablesize = casemax - casemin + 1;
		/*
		 *  Sort the case labels.
		 */
		qsort(casearray,casecount,sizeof(struct caseentry),casecmp);

		/*
		 *  Scan for duplicates in the case array, eliminating them and
		 *  printing warnings for them.
		 */
		cptr = cptr2 = casearray+1;
		lastval = casearray->value;
		for(i=casecount;--i;) {
			if(cptr->value == lastval) {
				cptr++;
				casecount--;
				errl(DUPLICATE_CASE,node);
			} else {
				if(cptr2 != cptr) *cptr2 = *cptr;
				lastval = cptr->value;
				cptr++;
				cptr2++;
			}
		}

		/*
		 *  Decide what kind of switch statement implementation to use.  If
		 *  the case labels "sparsely" populate the integer space, then
		 *  generate a binary dispatch tree of compare and branch instructions,
		 *  whose traversal time is O(log2(n)).  If they "densely" populate
		 *  a range of integers, then generate a CASE instruction, whose
		 *  execution time is constant.
		 *
		 *  The decision criterion is that when a CASE instruction is
		 *  generated, at least one in 10 table entries must point somewhere
		 *  other than the default, and there must be more than two entries.
		 */
		if(casecount > 2 && 10*casecount >= tablesize) {
			/*
			 *  Generate the switch expression and the case instruction.
			 */
			en1.type = en2.type = ES_CONST;
			en1.u.val1 = casemin;
			en2.u.val1 = tablesize-1;
			genexpr(node->link2,0,4);
			geninst("casel",3,&expr_stack[es_pointer],&en1,&en2);
			es_pop(1);
			baselabel = labelcnt++;
			genlabel(baselabel);
			/*
			 *  Generate the dispatch table.
			 */
			cptr = casearray;
			for(i=0;i<tablesize;i++) {
				if(casecount && cptr->value == i+casemin) {
					lab = cptr->label;
					cptr++;
					casecount--;
				} else lab = defaultlabel;
				printf("\t.word\tL%d-L%d\n",lab,baselabel);
			}
			/*
			 *  Generate the jump instruction used when the case instruction
			 *  "falls through" due to out-of-range argument.
			 */
			en1.type = ES_STATIC;
			en1.u.val1 = defaultlabel;
			en1.val2 = 0;
			geninst("jbr",1,&en1);
		} else {
			/*
			 *  Generate the slect expression for the binary tree case.
			 */
			genexpr(node->link2,0,4);
			registerize(4);
			ep = &expr_stack[es_pointer];
			en1.type = ES_CONST;
			en2.type = ES_STATIC;
			en2.val2 = 0;
			/*
			 *  Build the compare/branch tree.
			 */
			makebinarysearch(ep,&en1,&en2,casearray,
				casearray+casecount-1,defaultlabel);
			es_pop(1);
		}
		free(casearray);
	} else {
		/*
		 *  If no cases at all are given, then issue a warning but allow it.
		 */
		errl(EMPTYSWITCH);
		en1.type = ES_STATIC;
		en1.u.val1 = defaultlabel;
		en1.val2 = 0;
		geninst("jbr",1,&en1);
	}
	/*
	 *  Finally, generate code for the statement following the switch.
	 */
	codegen(node->link3,endlabel,contlabel);
	genlabel(endlabel);
}

/*
 *  This routine builds a "binary dispatch tree" of compare and branch
 *  instructions which can route control to one of n destinations in time
 *  O(log2(n)).  It is recursive, of course.
 *
 *  mode1, mode2:  Addressing mode entries for the "compare" instruction.
 *  mode3:         A spare addressing mode structure to use for the conditional
 *				   branches (preinitialized to ES_STATIC).
 *  first, last:   The highest and lowest case values being considered, given
 *				   as pointers to case array entries.
 *  endlabel:      The label to jump to when none of the case values were
 *				   matched.
 */
makebinarysearch(mode1,mode2,mode3,first,last,endlabel)
register struct es_entry *mode1,*mode2,*mode3;
register struct caseentry *first, *last;
{
	int lab = 0;
	struct caseentry *middle = first + (last-first)/2;
	mode2->u.val1 = middle->value;
	geninst("cmpl",2,mode1,mode2);
	mode3->u.val1 = middle->label;
	geninst("jeql",1,mode3);
	if(first==last) {
		mode3->u.val1 = endlabel;
		geninst("jbr",1,mode3);
	} else {
		if(middle!=first) {
			if(middle!=last) {
				lab = labelcnt++;
				mode3->u.val1 = lab;
				geninst("jgtr",1,mode3);
			}
			makebinarysearch(mode1,mode2,mode3,first,middle-1,endlabel);
			if(lab) genlabel(lab);
		}
		if(middle!=last)
			makebinarysearch(mode1,mode2,mode3,middle+1,last,endlabel);
	}
}

/*
 *  This routine traverses a statement node, gathering up any CASE and
 *  DEFAULT nodes which are not hidden by a lower-down SWITCH node.  The
 *  CASE nodes are strung together by a spare link entry; this turns the
 *  tree into something which is not a tree and therefore, care must be
 *  be taken not to deallocate or otherwise traverse it while the nodes
 *  are thus strung together.
 */
gathercases(node)
register struct treenode *node;
{
	register struct treenode *n1;
	switch(node->ntype) {
		case TN_LABEL:
			n1 = node->link2;
			if(n1->ntype == TN_CASE) {
				*current = n1;
				current = &n1->link3;
				*current = 0;
				n1->link2->format = NF_LEAF_STATIC;
				n1->link2->v_sval = labelcnt++;
				++casecount;
			} else if(n1->ntype == TN_DEFAULT) {
				if(defaultlabel) errl(DUPDEFAULT,n1);
				else {
					n1->link1->format = NF_LEAF_STATIC;
					defaultlabel = n1->link1->v_sval = labelcnt++;
				}
			}
			gathercases(node->link3);
			break;
		case TN_COMPOUNDSTAT:
			node = node->link2;
			while(node) {
				gathercases(node);
				node = node->link1;
			}
			break;
		case TN_IF:
			gathercases(node->link3);
			break;
		case TN_IF_ELSE:
			gathercases(node->link3->link1);
			gathercases(node->link3->link2);
			break;
			break;
		case TN_WHILE:
		case TN_FOR:
			gathercases(node->link3);
			break;
		case TN_DOWHILE:
			gathercases(node->link2);
			break;
		case TN_SWITCH:
		case TN_EXPR_STAT:
		case TN_GOTO:
		case TN_CONTINUE:
		case TN_BREAK:
		case TN_RETURN:
			break;
		default:
			panic("hit unknown node while gathering case labels");
	}
}
